# RCAIDE/Framework/Analyses/Aerodynamics/Vortex_Lattice_Method.py
#  
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports   
from RCAIDE.Framework.Core                             import Data, Units
from RCAIDE.Framework.Analyses                         import Process 
from RCAIDE.Library.Methods.Aerodynamics               import Common
from .Aerodynamics                                     import Aerodynamics 
from RCAIDE.Framework.Analyses.Common.Process_Geometry import Process_Geometry 
from RCAIDE.Library.Methods.Aerodynamics.Vortex_Lattice_Method import *   

# package imports 
import numpy as np 

# ----------------------------------------------------------------------------------------------------------------------
#  Vortex_Lattice_Method
# ---------------------------------------------------------------------------------------------------------------------- 
class Vortex_Lattice_Method(Aerodynamics):
    """This is a subsonic aerodynamic buildup analysis based on the vortex lattice method

     Assumptions:
     Stall effects are negligible 
 
     Source:
     N/A
 
     Inputs:
     None
 
     Outputs:
     None
 
     Properties Used:
     N/A 
    """      
    
    def __defaults__(self):
        """This sets the default values and methods for the analysis.

        Assumptions:
        None

        Source:
        N/A

        Inputs:
        None

        Outputs:
        None

        Properties Used:
        N/A
        """          
        self.tag                                                        = 'Vortex_Lattice_Method'  
        self.vehicle                                                    = Data()  
        self.process                                                    = Process()
        self.process.initialize                                         = Process()  
                   
        # correction factors            
        self.settings.fuselage_lift_correction                           = 1.14
        self.settings.trim_drag_correction_factor                        = 1.0
        self.settings.wing_parasite_drag_form_factor                     = 1.1
        self.settings.fuselage_parasite_drag_form_factor                 = 2.3
        self.settings.maximum_lift_coefficient_factor                    = 1.0        
        self.settings.lift_to_drag_adjustment                            = 0.0  
        self.settings.oswald_efficiency_factor                           = None
        self.settings.span_efficiency                                    = None
        self.settings.viscous_lift_dependent_drag_factor                 = 0.38
        self.settings.drag_coefficient_increment                         = 0.0
        self.settings.spoiler_drag_increment                             = 0.0
        self.settings.maximum_lift_coefficient                           = np.inf 
        self.settings.use_surrogate                                      = True
        self.settings.recalculate_total_wetted_area                      = False
        self.settings.propeller_wake_model                               = False 
        self.settings.discretize_control_surfaces                        = True
        self.settings.model_fuselage                                     = False
        self.settings.trim_aircraft                                      = True 
        self.settings.aileron_flag                                       = False
        self.settings.rudder_flag                                        = False
        self.settings.flap_flag                                          = False
        self.settings.elevator_flag                                      = False
        self.settings.slat_flag                                          = False             

        # correction factors
        self.settings.supersonic                                         = Data()
        self.settings.supersonic.peak_mach_number                        = 1.04  
        self.settings.supersonic.begin_drag_rise_mach_number             = 0.95
        self.settings.supersonic.end_drag_rise_mach_number               = 1.2
        self.settings.supersonic.transonic_drag_multiplier               = 1.25  
        self.settings.supersonic.volume_wave_drag_scaling                = 3.2  
        self.settings.supersonic.fuselage_parasite_drag_begin_blend_mach = 0.91
        self.settings.supersonic.fuselage_parasite_drag_end_blend_mach   = 0.99    
        self.settings.supersonic.cross_sectional_area_calculation_type   = 'Fixed'     
        self.settings.supersonic.wave_drag_type                          = 'Raymer'    
    
        self.settings.number_of_spanwise_vortices                        = 15
        self.settings.number_of_chordwise_vortices                       = 5
        self.settings.wing_spanwise_vortices                             = None
        self.settings.wing_chordwise_vortices                            = None
        self.settings.fuselage_spanwise_vortices                         = None
        self.settings.fuselage_chordwise_vortices                        = None  
        self.settings.spanwise_cosine_spacing                            = True
        self.settings.vortex_distribution                                = Data()  
        self.settings.leading_edge_suction_multiplier                    = 1.0  
        self.settings.use_VORLAX_matrix_calculation                      = False
        self.settings.floating_point_precision                           = np.float32     
    
        # conditions table, used for surrogate model training
        self.training                                               = Data()
        self.training.angle_of_attack                               = np.array([-5., -2. , 1E-20 , 2.0, 5.0, 8.0, 12., 45., 75.]) * Units.deg 
        self.training.Mach                                          = np.array([0.1  ,0.3,  0.5,  0.65 , 0.85 , 0.9, 1.3, 1.35 , 1.5 , 2.0, 2.25 , 2.5  , 3.5])             
                      
        self.training.subsonic                                      = None
        self.training.supersonic                                    = None
        self.training.transonic                                     = None

        self.training.altitude                                      = 0
        self.training.speed_of_sound                                = 343 
        self.training.angle_purtubation                             = 2 * Units.deg          
        self.training.speed_purtubation                             = 0.5  
        self.training.rate_purtubation                              = 0.5 * Units.deg / Units.sec   
        self.training.control_surface_purtubation                   = 5 * Units.deg   
        self.training.sideslip_angle                                = np.array([10  , 5.0 ]) * Units.deg
        self.training.aileron_deflection                            = np.array([10  , 5.0 ]) * Units.deg
        self.training.elevator_deflection                           = np.array([10  , 5.0 ]) * Units.deg   
        self.training.rudder_deflection                             = np.array([10  , 1E-3 ]) * Units.deg
        self.training.flap_deflection                               = np.array([10  , 1E-3 ]) * Units.deg 
        self.training.slat_deflection                               = np.array([10  , 1E-3 ]) * Units.deg                      
        self.training.u                                             = np.array([10 , 5 ])  
        self.training.v                                             = np.array([10 , 5 ])  
        self.training.w                                             = np.array([10 , 5 ])    
        self.training.pitch_rate                                    = np.array([3 ,1.5 ])  * Units.deg / Units.sec
        self.training.roll_rate                                     = np.array([3 ,1.5 ])  * Units.deg / Units.sec
        self.training.yaw_rate                                      = np.array([3 ,1.5 ])  * Units.deg / Units.sec
    
        self.reference_values                                       = Data()
        self.reference_values.S_ref                                 = 0
        self.reference_values.c_ref                                 = 0
        self.reference_values.b_ref                                 = 0
        self.reference_values.X_ref                                 = 0
        self.reference_values.Y_ref                                 = 0
        self.reference_values.Z_ref                                 = 0
        
        # control surface flags                  
        self.aileron_flag                                           = False 
        self.flap_flag                                              = False 
        self.rudder_flag                                            = False 
        self.elevator_flag                                          = False 
        self.slat_flag                                              = False
                         
        # blending function                  
        self.hsub_min                                               = 0.85
        self.hsub_max                                               = 0.95
        self.hsup_min                                               = 1.05
        self.hsup_max                                               = 1.15 
                     
        # surrogoate models                 
        self.surrogates                                             = Data() 

        # build the evaluation process
        compute                                    = Process() 
        compute.lift                               = Process() 
        compute.lift.inviscid_wings                = None 
        compute.lift.fuselage                      = Common.Lift.fuselage_correction 
        compute.drag                               = Process()
        compute.drag.parasite                      = Process()
        compute.drag.parasite.wings                = Process_Geometry('wings')
        compute.drag.parasite.wings.wing           = Common.Drag.parasite_drag_wing 
        compute.drag.parasite.fuselages            = Process_Geometry('fuselages')
        compute.drag.parasite.fuselages.fuselage   = Common.Drag.parasite_drag_fuselage
        compute.drag.parasite.booms                = Process_Geometry('booms')
        compute.drag.parasite.booms.boom           = Common.Drag.parasite_drag_fuselage 
        compute.drag.parasite.nacelles             = Common.Drag.parasite_drag_nacelle
        compute.drag.parasite.pylons               = Common.Drag.parasite_drag_pylon
        compute.drag.parasite.total                = Common.Drag.parasite_total
        compute.drag.induced                       = Common.Drag.induced_drag
        compute.drag.cooling                       = Process()
        compute.drag.cooling.total                 = Common.Drag.cooling_drag        
        compute.drag.compressibility               = Process() 
        compute.drag.compressibility.total         = Common.Drag.compressibility_drag
        compute.drag.miscellaneous                 = Common.Drag.miscellaneous_drag 
        compute.drag.spoiler                       = Common.Drag.spoiler_drag
        compute.drag.total                         = Common.Drag.total_drag
        compute.stability                          = Process()
        compute.stability.dynamic_modes            = RCAIDE.Library.Methods.Stability.Common.compute_dynamic_flight_modes  
        self.process.compute                       = compute
        

    def initialize(self):  
        use_surrogate   = self.settings.use_surrogate  

        # If we are using the surrogate
        if use_surrogate == True: 
            # sample training data
            train_VLM_surrogates(self)

            # build surrogate
            build_VLM_surrogates(self)  
    
        # build the evaluation process
        compute   =  self.process.compute                  
        if use_surrogate == True: 
            compute.lift.inviscid_wings  = evaluate_surrogate
        else:
            compute.lift.inviscid_wings  = evaluate_no_surrogate  
        return 
    
         
    def evaluate(self,state):
        """The default evaluate function.

        Assumptions:
        None

        Source:
        N/A

        Inputs:
        None

        Outputs:
        results   <RCAIDE data class>

        Properties Used:
        self.settings
        self.vehicle
        """          
        settings = self.settings
        vehicle  = self.vehicle 
        results  = self.process.compute(state,settings,vehicle)
        
        return results


# RCAIDE/Library/Compoments/Thermal_Management/Accessories/Fan.py
# 
#
# Created: March 2024  S. Shekar

# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------
from RCAIDE.Framework.Core import Data

# ----------------------------------------------------------------------
#  Fan
# ----------------------------------------------------------------------
class Fan(Data):
    """
    A class representing a cooling fan for thermal management systems.

    Attributes
    ----------
    tag : str
        Unique identifier for the fan component, defaults to 'Fan'
        
    efficiency : float
        Overall fan efficiency including mechanical and electrical losses, 
        defaults to 1.0

    Notes
    -----
    The fan class models forced air movement devices used in thermal management 
    systems. It provides functionality for:
    
    * Power consumption calculation
    * Performance modeling based on operating conditions
    * Integration with cooling system analysis

    **Definitions**

    'Pressure Differential'
        The difference in pressure across the fan
        
    'Mass Flow Rate'
        Rate of air mass flow through the fan

    See Also
    --------
    RCAIDE.Library.Components.Thermal_Management.Accessories.Pump
        Similar component for liquid cooling systems
    """

    def __defaults__(self):
        """
        Sets default values for the fan attributes.
        """
        self.tag        = 'Fan'
        self.efficiency = 1.0
        return
   
    def compute_power_consumed(pressure_differential, density, mass_flow_rate, efficiency):
        """
        Calculates the power consumed by the fan.

        Parameters
        ----------
        pressure_differential : float
            Pressure rise across the fan
            
        density : float
            Fluid density
            
        mass_flow_rate : float
            Mass flow rate through the fan
            
        efficiency : float
            Overall fan efficiency

        Returns
        -------
        float
            Power consumed by the fan

        Notes
        -----
        Uses the standard fan power equation:
        Power = (mass_flow_rate * pressure_differential) / (density * efficiency)
        """
        return mass_flow_rate * pressure_differential / (density * efficiency)


# RCAIDE/Library/Methods/Emissions/Chemical_Reactor_Network_Method/__init__.py

# Created: June 2024, M. Clarke, M. Guidotti 

"""
Chemical Reactor Network (CRN) Method 

This module provides functionality for building, training and evaluating surrogate models as well 
as for direct computation of emission indices using Chemical Reactor Networks.

Notes
-----
The Chemical Reactor Network Method package contains tools for:

- Training CRN emission index surrogate models
- Building CRN emission index surrogate models
- Evaluating chemical reactions using Cantera
- Computing emission indices using CRN models

See Also
--------
RCAIDE.Library.Methods.Emissions.Chemical_Reactor_Network_Method.train_CRN_EI_surrogates
RCAIDE.Library.Methods.Emissions.Chemical_Reactor_Network_Method.build_CRN_EI_surrogates
RCAIDE.Library.Methods.Emissions.Chemical_Reactor_Network_Method.evaluate_CRN_emission_indices
RCAIDE.Library.Methods.Emissions.Chemical_Reactor_Network_Method.evaluate_cantera
RCAIDE.Library.Components.Propulsors.Converters.Combustor
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
 
from .build_CRN_EI_surrogates       import build_CRN_EI_surrogates
from .train_CRN_EI_surrogates       import train_CRN_EI_surrogates
from .evaluate_cantera              import evaluate_cantera
from .evaluate_CRN_emission_indices import *  


# RCAIDE/Framework/Analyses/Mission/Segments/Ground/Landing.py
# 
# 
# Created:  Jul 2023, M. Clarke
 
# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports
from RCAIDE.Framework.Mission.Segments.Evaluate        import Evaluate 
from RCAIDE.Framework.Core                                      import Units , Data 
from RCAIDE.Library.Mission.Segments                  import Ground  
from RCAIDE.Library.Mission.Common                    import Residuals , Unpack_Unknowns, Update

# ----------------------------------------------------------------------------------------------------------------------
#  Landing
# ----------------------------------------------------------------------------------------------------------------------

class Landing(Evaluate):
    """ Segment for landing. Integrates equations of motion
        including rolling friction.
        
        Assumptions:
        Notes Regarding Friction Coefficients
        Dry asphalt or concrete: .04 brakes off, .4 brakes on
        Wet asphalt or concrete: .05 brakes off, .225 brakes on
        Icy asphalt or concrete: .02 brakes off, .08 brakes on
        Hard turf:               .05 brakes off, .4 brakes on
        Firm dirt:               .04 brakes off, .3 brakes on
        Soft turf:               .07 brakes off, .2 brakes on
        Wet grass:               .08 brakes off, .2 brakes on
        
        Source: General Aviation Aircraft Design: Applied Methods and Procedures,
        by Snorri Gudmundsson, copyright 2014, published by Elsevier, Waltham,
        MA, USA [p.938]
    """     

    def __defaults__(self):
        """ This sets the default solver flow. Anything in here can be modified after initializing a segment.
    
            Assumptions:
            None
    
            Source:
            N/A
    
            Inputs:
            None
    
            Outputs:
            None
    
            Properties Used:
            None
        """            
        # -------------------------------------------------------------------------------------------------------------- 
        #   User Inputs
        # -------------------------------------------------------------------------------------------------------------- 
 
        self.velocity_start       = 150 * Units.knots
        self.velocity_end         = 0.0
        self.friction_coefficient = 0.4
        self.throttle             = 0.1
        self.altitude             = 0.0
        self.reverse_thrust_ratio = 0.1
        self.true_course          = 0.0 * Units.degrees 
        
        # -------------------------------------------------------------------------------------------------------------- 
        #  Unique Mission Unknowns and Residuals
        # -------------------------------------------------------------------------------------------------------------- 
        ones_row_m1                               = self.state.ones_row_m1
        self.state.residuals.final_velocity_error = 0.0
        self.state.residuals.force_x              = ones_row_m1(1) * 0.0    
        self.state.unknowns.elapsed_time          = 30.                        
        self.state.unknowns.ground_velocity       = ones_row_m1(1) * 0  

        # -------------------------------------------------------------------------------------------------------------- 
        #  Mission Conditions 
        # --------------------------------------------------------------------------------------------------------------          
        ones_row = self.state.ones_row  
        self.state.conditions.ground                              = Data() 
        self.state.conditions.ground.friction_coefficient         = ones_row(1) * 0.0
        self.state.conditions.frames.inertial.ground_force_vector = ones_row(3) * 0.0 
        
        # -------------------------------------------------------------------------------------------------------------- 
        #  Mission specific processes 
        # --------------------------------------------------------------------------------------------------------------  
        initialize                         = self.process.initialize
        initialize.conditions              = Ground.Landing.initialize_conditions  
        iterate                            = self.process.iterate   
        iterate.conditions.forces_ground   = Update.ground_forces
        iterate.unknowns.mission           = Unpack_Unknowns.ground
        iterate.residuals.flight_dynamics  = Residuals.flight_dynamics    

        return


# RCAIDE/Library/Methods/Emissions/Emission_Index_Empirical_Method/__init__.py

# Created: 2024, M. Clarke

"""
Emission Index Empirical Method 

This module provides functionality for evaluating emission indices using empirical correlations.

Functions
---------
evaluate_correlation_emissions_indices
    Computes emission indices using empirical correlations 

**Major Assumptions**

* Correlations are valid within the operating range

See Also
--------
RCAIDE.Library.Methods.Emissions.Emission_Index_Empirical_Method.evaluate_correlation_emissions_indices

"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .evaluate_correlation_emissions_indices  import evaluate_correlation_emissions_indices


# RCAIDE/Methods/Aerodynamics/Airfoil_Panel_Method/thwaites_method_new.py
# 
# 
# Created:  Apr 2024, Niranjan Nanjappa

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE imports  
from RCAIDE.Framework.Core import Data 

# pacakge imports  
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
#  Thwaites Method
# ----------------------------------------------------------------------------------------------------------------------
def thwaites_method(npanel,ncases,ncpts,NU,L,RE_L,X_I,VE_I, DVE_I,tol,THETA_0):
    """ Computes the boundary layer characteristics in laminar 
    flow pressure gradients
    
    Source:
    Thwaites, Bryan. "Approximate calculation of the laminar boundary layer." 
    Aeronautical Quarterly 1.3 (1949): 245-280.
    
    Assumptions:
    None  

    Inputs:  
    npanel         - number of points on surface                                                 [unitless]
    ncases         - number of cases                                                             [unitless]
    ncpts          - number of control points                                                    [unitless]
    batch_analysis - flag for batch analysis                                                     [boolean]
    THETA_0        - initial momentum thickness                                                  [m]
    L              - normalized length of surface                                                [unitless]
    RE_L           - Reynolds number                                                             [unitless]
    X_I            - x coordinate on surface of airfoil                                          [unitless]
    VE_I           - boundary layer velocity at transition location                              [m/s] 
    DVE_I          - initial derivative value of boundary layer velocity at transition location  [m/s-m] 
    tol            - boundary layer error correction tolerance                                   [unitless]

    Outputs: 
    RESULTS.
      X_T          - reshaped distance along airfoil surface             [unitless]
      THETA_T      - momentum thickness                                  [m]
      DELTA_STAR_T - displacement thickness                              [m] 
      H_T          - shape factor                                        [unitless]
      CF_T         - friction coefficient                                [unitless]
      RE_THETA_T   - Reynolds number as a function of momentum thickness [unitless]
      RE_X_T       - Reynolds number as a function of distance           [unitless]
      DELTA_T      - boundary layer thickness                            [m]

    Properties Used:
    N/A
    """ 
    # Initialize vectors
    X_T          = np.zeros((npanel,ncases,ncpts))
    THETA_T      = np.zeros_like(X_T)
    DELTA_STAR_T = np.zeros_like(X_T)
    H_T          = np.zeros_like(X_T)
    CF_T         = np.zeros_like(X_T)
    RE_THETA_T   = np.zeros_like(X_T)
    RE_X_T       = np.zeros_like(X_T)
    DELTA_T      = np.zeros_like(X_T)  
      
    for case in range(ncases):
        for cpt in range(ncpts):
            
            def dy_by_dx(index, X, Y):
                return 0.45*nu*Ve_i[index]**5            
            
            l              = L[case,cpt]
            theta_0        = THETA_0 
            Re_L           = RE_L[case,cpt]
            x_i            = X_I.data[:,case,cpt][X_I.mask[:,case,cpt] ==False]
            Ve_i           = VE_I.data[:,case,cpt][VE_I.mask[:,case,cpt] ==False]
            dVe_i          = DVE_I.data[:,case,cpt][DVE_I.mask[:,case,cpt] ==False]
            nu             = NU[case,cpt]
            n              = len(x_i)
            dx_i           = np.diff(x_i)
            theta2_Ve6     = np.zeros(n)
            theta2_Ve6[0]  = (theta_0**2)*Ve_i[0]**6
            
            # determine (Theta**2)*(Ve**6)
            for i in range(1,n):
                theta2_Ve6[i] = RK4(i-1, dx_i, x_i, theta2_Ve6, dy_by_dx)
            
            # Compute momentum thickness
            theta       = np.sqrt(theta2_Ve6/Ve_i**6)
            
            # find theta values that do not converge and replace them with neighbor
            idx1        = np.where(abs((theta[1:] - theta[:-1])/theta[:-1]) > tol)[0] 
            if len(idx1)> 1:  
                np.put(theta,idx1 + 1, theta[idx1])
                
            # Thwaites separation criteria 
            lambda_val  = theta**2*dVe_i/nu 
            
            # Compute H 
            H           = getH(lambda_val)
            H[H<0]      = 1E-6   # H cannot be negative 
            # find H values that do not converge and replace them with neighbor
            idx1        = np.where(abs((H[1:] - H[:-1])/H[:-1]) > tol)[0]
            if len(idx1)> 1: 
                np.put(H,idx1 + 1, H[idx1]) 
            
            # Compute Reynolds numbers based on momentum thickness  
            Re_theta    = Ve_i*theta/nu
            
            # Compute Reynolds numbers based on distance along airfoil
            Re_x        = Ve_i*x_i/nu
            
            # Compute skin friction 
            cf          = abs(getcf(lambda_val, Re_theta)) 
            
            # Compute displacement thickness
            del_star    = H*theta   
            
            # Compute boundary layer thickness 
            delta       = 5.2*x_i/np.sqrt(Re_x)
            delta[0]    = 0   
            
            # Reynolds number at x=0 cannot be negative 
            Re_x[0]     = 1E-5
            
            # Find where matrices are not masked 
            indices = np.where(X_I.mask[:,case,cpt] == False)
            
            # Store results 
            np.put(X_T[:,case,cpt],indices,x_i)
            np.put(THETA_T[:,case,cpt],indices,theta)
            np.put(DELTA_STAR_T[:,case,cpt],indices,del_star)
            np.put(H_T[:,case,cpt],indices,H)
            np.put(CF_T[:,case,cpt],indices ,cf)
            np.put(RE_THETA_T[:,case,cpt],indices,Re_theta)
            np.put(RE_X_T[:,case,cpt],indices,Re_x)
            np.put(DELTA_T[:,case,cpt],indices,delta)
    
    RESULTS = Data(
        X_T          = X_T,      
        THETA_T      = THETA_T,   
        DELTA_STAR_T = DELTA_STAR_T,
        H_T          = H_T,       
        CF_T         = CF_T,      
        RE_THETA_T   = RE_THETA_T,   
        RE_X_T       = RE_X_T,    
        DELTA_T      = DELTA_T,  
    )    
    
    return RESULTS
            



def getH(lambda_val ): 
    """ Computes the shape factor, H

    Assumptions:
    None

    Source:
    None

    Inputs: 
    lamdda_val  - thwaites separation criteria [unitless]

    Outputs:  
    H           - shape factor [unitless]

    Properties Used:
    N/A
    """       
    H       = 0.0731/(0.14 + lambda_val ) + 2.088 
    idx1    = (lambda_val>0.0)  
    H[idx1] = 2.61 - 3.75*lambda_val[idx1]  + 5.24*lambda_val[idx1]**2   
    return H


def getcf(lambda_val , Re_theta):
    """ Computes the skin friction coefficient, cf

    Assumptions:
    None

    Source:
    None

    Inputs: 
    lambda_val - thwaites separation criteria                        [unitless]
    Re_theta   - Reynolds Number as a function of momentum thickness [unitless]

    Outputs:  
    cf         - skin friction coefficient [unitless]

    Properties Used:
    N/A 
    """        
    l       = 0.22 + 1.402*lambda_val  + (0.018*lambda_val)/(0.107 + lambda_val ) 
    idx1    = (lambda_val>0.0)   
    l[idx1] = 0.22 + 1.57*lambda_val[idx1] - 1.8*lambda_val[idx1]**2 
    cf      = 2*l/Re_theta  
    return cf


def RK4(ind, dx, x, Var, Slope):
    m1 = Slope(ind,  x[ind],  Var[ind])
    m2 = Slope(ind,  x[ind] + dx[ind]/2,  Var[ind] + m1*dx[ind]/2)
    m3 = Slope(ind,  x[ind] + dx[ind]/2,  Var[ind] + m2*dx[ind]/2)
    m4 = Slope(ind,  x[ind] + dx[ind],  Var[ind] + m3*dx[ind])
    
    change = (dx[ind]/6)*(m1 + 2*m2 + 2*m3 + m4)
    return Var[ind] + change            
            



# RCAIDE/Library/Components/Fuselages/__init__.py

"""
Module containing fuselage components for aircraft design and analysis. This module provides 
different fuselage configurations including traditional tube fuselages, blended wing body 
designs.

See Also
--------
RCAIDE.Library.Components.Airfoils
    Related module for airfoil definitions used in aerodynamic surfaces
RCAIDE.Library.Components.Booms
    Related module for boom components that may connect to fuselage sections
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .Fuselage                   import Fuselage
from .Blended_Wing_Body_Fuselage import Blended_Wing_Body_Fuselage 
from .Tube_Fuselage              import Tube_Fuselage

from . import Segments


# RCAIDE/Library/Methods/Propulsor/Ducted_Fan_Propulsor/run_dfdc_analysis.py
# 
# Created: Sep 2024, M. Clarke 

# ---------------------------------------------------------------------------------------------------------------------- 
#  Imports
# ----------------------------------------------------------------------------------------------------------------------

from RCAIDE.Framework.Core  import redirect 
import sys 
import subprocess
import os 
from .purge_files  import purge_files 

# ---------------------------------------------------------------------------------------------------------------------- 
# Run DFDC Analysis
# ----------------------------------------------------------------------------------------------------------------------   
def run_dfdc_analysis(dfdc_object,print_output):
    """ This calls the DFDC executable and runs an analysis

    Assumptions:
        None
        
    Source:
        None

    Inputs:
        dfdc_object - passed into the  call_dfdc function  
        
    Outputs:
        results

    Properties Used:
        N/A
    """      
    new_regression_results = dfdc_object.settings.new_regression_results
    if new_regression_results:
        exit_status = 0 
    else:
        log_file = dfdc_object.settings.filenames.log_filename
        err_file = dfdc_object.settings.filenames.err_filename
        if isinstance(log_file,str):
            purge_files(log_file)
        if isinstance(err_file,str):
            purge_files(err_file)
        dfdc_call = dfdc_object.settings.filenames.dfdc_bin_name
        case      = dfdc_object.settings.filenames.case
        in_deck   = dfdc_object.current_status.deck_file  
    
        with redirect.output(log_file,err_file): 
            with open(in_deck,'r') as commands: 
                
                # Initialize suppression of console window output
                if print_output == False:
                    devnull = open(os.devnull,'w')
                    sys.stdout = devnull       
                    
                # Run DFDC
                dfdc_run = subprocess.Popen([dfdc_call,case],stdout=sys.stdout,stderr=sys.stderr,stdin=subprocess.PIPE)
                for line in commands:
                    dfdc_run.stdin.write(line.encode('utf-8'))
                    dfdc_run.stdin.flush()
                    
                  
                # Terminate suppression of console window output  
                if print_output == False:
                    sys.stdout = sys.__stdout__                    
                    
            dfdc_run.wait()
    
            exit_status = dfdc_run.returncode 

    return exit_status




# RCAIDE/Library/Compoments/Thermal_Management/Accessories/__init__.py 

"""
Module containing accessory components for thermal management systems. This module provides 
classes for fluid movement and thermal control devices including fans and pumps.

See Also
--------
RCAIDE.Library.Components.Thermal_Management.Heat_Exchangers
    Related module for heat transfer components that work with these accessories
RCAIDE.Library.Components.Thermal_Management.Reservoirs
    Related module for fluid storage components in thermal systems
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
from .Fan  import Fan 
from .Pump import Pump




# RCAIDE/Library/Attributes/Propellants/Jet_A1.py
# 
#
# Created:  Mar 2024, M. Clarke

# ---------------------------------------------------------------------------------------------------------------------- 
#  Imports
# ---------------------------------------------------------------------------------------------------------------------- 

from .Propellant import Propellant
from RCAIDE.Framework.Core import Data  

# ---------------------------------------------------------------------------------------------------------------------- 
#  Jet_A1 Propellant Class
# ----------------------------------------------------------------------------------------------------------------------  
class Jet_A1(Propellant):
    """Jet A1 class propellant  
    """

    def __defaults__(self):
        """This sets the default values.
    
        Assumptions:
            None
        
        Source:
            lower_heating_value: Boehm et al, Lower Heating Value of Jet Fuel From Hydrocarbon Class Concentration Data
            and Thermo-Chemical Reference Data: An Uncertainty Quantification
            
            emission indices: NASA's Engine Performance Program (NEPP) and 
    
        """    
        self.tag                       = 'Jet A1'
        self.reactant                  = 'O2'
        self.density                   = 804.0                            # kg/m^3 (15 C, 1 atm)
        self.specific_energy           = 43.15e6                          # J/kg
        self.energy_density            = 34692.6e6                        # J/m^3
        self.lower_heating_value       = 43.24e6                          # J/kg 
        self.max_mass_fraction         = Data({'Air' : 0.0633, 'O2' : 0.3022})  # kg propellant / kg oxidizer
        self.temperatures.flash        = 311.15                           # K
        self.temperatures.autoignition = 483.15                           # K
        self.temperatures.freeze       = 226.15                           # K
        self.temperatures.boiling      = 0.0                              # K 

   
        self.use_high_fidelity_kinetics_model      =  True 
        self.fuel_surrogate_chemical_properties    = {'N-C12H26':0.6, 'A1CH3':0.2, 'A1':0.2}
        self.fuel_chemical_properties              = {'NC10H22':0.16449, 'NC12H26':0.34308, 'NC16H34':0.10335, 'IC8H18':0.08630, 'NC7H14':0.07945, 'C6H5C2H5': 0.07348, 'C6H5C4H9': 0.05812, 'C10H7CH3': 0.10972}      # [2] More accurate kinetic mechanism, slower simulation    
        self.air_chemical_properties               = {'O2':0.2095, 'N2':0.7809, 'AR':0.0096}
        self.surrogate_species_list                = ['CO', 'CO2', 'H2O']
        self.species_list                          = ['CO', 'CO2', 'H2O', 'NO', 'NO2', 'CSOLID']
        self.surrogate_chemical_kinetics           = 'Fuel_Surrogate.yaml'
        self.chemical_kinetics                     = 'Fuel.yaml'
        self.oxidizer                              = 'Air.yaml'
        
        # critical temperatures   
        self.temperatures.flash           = 311.15                 # K
        self.temperatures.autoignition    = 483.15                 # K
        self.temperatures.freeze          = 233.15                 # K
        self.temperatures.boiling         = 0.0                    # K  
        

        self.emission_indices.Production  = 0.4656   # kg/kg Greet 
        self.emission_indices.CO2         = 3.16    # kg/kg  fuel
        self.emission_indices.H2O         = 1.34    # kg/kg  fuel 
        self.emission_indices.SO2         = 0.0012  # kg/kg  fuel
        self.emission_indices.NOx         = 0.01514 # kg/kg  fuel
        self.emission_indices.Soot        = 0.0012  # kg/kg  fuel

        self.global_warming_potential_100.CO2       = 1     # CO2e/kg  
        self.global_warming_potential_100.H2O       = 0.06  # CO2e/kg  
        self.global_warming_potential_100.SO2       = -226  # CO2e/kg  
        self.global_warming_potential_100.NOx       = 52    # CO2e/kg  
        self.global_warming_potential_100.Soot      = 1166  # CO2e/kg    
        self.global_warming_potential_100.Contrails = 11 # kg/CO2e/km  


# RCAIDE/Library/Methods/Performance/compute_payload_range_diagram.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports
import RCAIDE
from RCAIDE.Framework.Core import Units , Data  
from RCAIDE.Library.Plots.Common import set_axes, plot_style    
 
# Pacakge imports 
import numpy as np
from matplotlib import pyplot as plt
 
# ----------------------------------------------------------------------
#  Calculate vehicle Payload Range Diagram
# ----------------------------------------------------------------------  
def compute_payload_range_diagram(vehicle,assigned_propulsors,weights_analysis,aerodynamics_analysis,cruise_airspeed = 515*Units.mph, cruise_altitude= 35000*Units.feet,max_range_guess = 1000*Units.nmi, reserves=0., plot_diagram = True, fuel_name=None):  
    """Calculates and plots the payload range diagram for an aircraft by modifying the cruise segment and weights of the aicraft .

        Sources:
        N/A

        Assumptions:
        None 

        Inputs:
            vehicle             data structure for aircraft                  [-]
            mission             data structure for mission                   [-] 
            cruise_segment_tag  string of cruise segment                     [string]
            reserves            reserve fuel                                 [unitless] 
            
        Outputs: 
            payload_range       data structure of payload range properties   [m/s]
    """ 

    configs  =  configs_setup(vehicle)

    # create analyses
    analyses =  analyses_setup(configs,weights_analysis,aerodynamics_analysis)              
    
    for network in vehicle.networks:
        if type(network) == RCAIDE.Framework.Networks.Fuel:
            electric_aircraft =  False
            
            # mission analyses 
            mission =  mission_setup(analyses,electric_aircraft,assigned_propulsors,cruise_airspeed, cruise_altitude, max_range_guess)
            
            # create mission instances (for multiple types of missions)
            missions = missions_setup(mission) 
            cruise_segment_tag = 'cruise'
            payload_range  =  conventional_payload_range_diagram(vehicle, missions.base_mission,cruise_segment_tag,reserves,plot_diagram,fuel_name) 
        elif type(network) == RCAIDE.Framework.Networks.Electric:
            electric_aircraft =  True
             
            # mission analyses 
            mission =  mission_setup(analyses,electric_aircraft,assigned_propulsors,cruise_airspeed, cruise_altitude, max_range_guess)
            
            # create mission instances (for multiple types of missions)
            missions = missions_setup(mission) 
            cruise_segment_tag = 'cruise'
            payload_range  =  electric_payload_range_diagram(vehicle,missions.base_mission,cruise_segment_tag,plot_diagram)
    return payload_range 
             
def conventional_payload_range_diagram(vehicle,mission,cruise_segment_tag,reserves,plot_diagram, fuel_name): 
    """Calculates and plots the payload range diagram for a fuel-bases aircraft by modifying the
    cruise segment range and weights of the aicraft .

        Sources:
        N/A

        Assumptions:
        None 

        Inputs:
            vehicle             data structure for aircraft                  [-]
            mission             data structure for mission                   [-] 
            cruise_segment_tag  string of cruise segment                     [string]
            reserves            reserve fuel                                 [unitless] 
            
        Outputs: 
            payload_range       data structure of payload range properties   [m/s]
    """ 
    #unpack
    mass = vehicle.mass_properties
    if not mass.operating_empty:
        raise AttributeError("Error calculating Payload Range Diagram: Vehicle Operating Empty not defined") 
    else:
        OEW = mass.operating_empty

    if not mass.max_zero_fuel:
        raise AttributeError("Error calculating Payload Range Diagram: Vehicle MZFW not defined") 
    else:
        MZFW = vehicle.mass_properties.max_zero_fuel

    if not mass.max_takeoff:
        raise AttributeError("Error calculating Payload Range Diagram: Vehicle MTOW not defined") 
    else:
        MTOW = vehicle.mass_properties.max_takeoff

    if mass.max_payload == 0:
        MaxPLD = MZFW - OEW  
    else:
        MaxPLD = vehicle.mass_properties.max_payload
        MaxPLD = min(MaxPLD , MZFW - OEW) #limit in structural capability

    if mass.max_fuel == 0:
        MaxFuel = MTOW - OEW # If not defined, calculate based in design weights
    else:
        MaxFuel = vehicle.mass_properties.max_fuel  # If max fuel capacity not defined
        MaxFuel = min(MaxFuel, MTOW - OEW)


    # Define payload range points
    #Point  = [ RANGE WITH MAX. PLD   , RANGE WITH MAX. FUEL , FERRY RANGE   ]
    TOW     = [ MTOW                               , MTOW                   , OEW + MaxFuel ]
    FUEL    = [ min(TOW[1] - OEW - MaxPLD,MaxFuel) , MaxFuel                , MaxFuel       ]
    PLD     = [ MaxPLD                             , min(MTOW - MaxFuel - OEW, MaxPLD)   , 0.   ]
    
    # allocating Range array
    R       = [0,0,0]

    # loop for each point of Payload Range Diagram
    for i in range(len(TOW)):
        ##    for i in [2]: 
        # Define takeoff weight
        mission.segments[0].analyses.weights.vehicle.mass_properties.takeoff = TOW[i]

        # Evaluate mission with current TOW
        results = mission.evaluate()
        segment = results.segments[cruise_segment_tag]

        # Distance convergency in order to have total fuel equal to target fuel
        #
        # User don't have the option of run a mission for a given fuel. So, we
        # have to iterate distance in order to have total fuel equal to target fuel
        #

        maxIter = 10    # maximum iteration limit
        tol     = 1.    # fuel convergency tolerance
        err     = 9999. # error to be minimized
        iter    = 0     # iteration count

        while abs(err) > tol and iter < maxIter:
            iter = iter + 1

            # Current total fuel burned in mission
            TotalFuel  = TOW[i] - results.segments[-1].conditions.weights.total_mass[-1,0]

            # Difference between burned fuel and target fuel
            missingFuel = FUEL[i] - TotalFuel - reserves

            # Current distance and fuel consuption in the cruise segment
            CruiseDist = np.diff( segment.conditions.frames.inertial.position_vector[[0,-1],0] )[0]        # Distance [m]
            CruiseFuel = segment.conditions.weights.total_mass[0,0] - segment.conditions.weights.total_mass[-1,0]    # [kg]
            # Current specific range (m/kg)
            CruiseSR    = CruiseDist / CruiseFuel        # [m/kg]

            # Estimated distance that will result in total fuel burn = target fuel
            DeltaDist  =  CruiseSR *  missingFuel
            mission.segments[cruise_segment_tag].distance = (CruiseDist + DeltaDist)

            # running mission with new distance
            results = mission.evaluate()
            segment = results.segments[cruise_segment_tag]

            # Difference between burned fuel and target fuel
            err = ( TOW[i] - results.segments[-1].conditions.weights.total_mass[-1,0] ) - FUEL[i] + reserves 

        # Allocating resulting range in ouput array.
        R[i] =  results.segments[-1].conditions.frames.inertial.position_vector[-1,0]   

    # Inserting point (0,0) in output arrays
    R.insert(0,0)
    PLD.insert(0,MaxPLD)
    FUEL.insert(0,0)
    TOW.insert(0,0)

    # packing results
    payload_range                = Data()
    payload_range.range          = np.array(R)
    payload_range.payload        = np.array(PLD)
    payload_range.fuel           = np.array(FUEL)
    payload_range.takeoff_weight = np.array(TOW)
    payload_range.reserves       = reserves
     
    if plot_diagram:  
        # get plotting style 
        ps      = plot_style()  
    
        parameters = {'axes.labelsize': ps.axis_font_size,
                      'xtick.labelsize': ps.axis_font_size,
                      'ytick.labelsize': ps.axis_font_size,
                      'axes.titlesize': ps.title_font_size}
        plt.rcParams.update(parameters)
        
        if fuel_name ==  None: 
            fig  = plt.figure('Fuel_Payload_Range_Diagram')
        else:
            fig  = plt.figure('Fuel_Payload_Range_Diagram for ' + fuel_name)
        axis = fig.add_subplot(1,1,1)
        axis.plot(payload_range.range /Units.nmi,payload_range.payload,color = 'k', linewidth = ps.line_width )
        axis.set_xlabel('Range (nautical miles)')
        axis.set_ylabel('Payload (kg)')
        axis.set_title("Fuel Payload Range Diagram") 
        fig.tight_layout()
        set_axes(axis) 

    return payload_range 
 
def electric_payload_range_diagram(vehicle,mission,cruise_segment_tag,plot_diagram):
    """Calculates and plots the payload range diagram for an electric aircraft by modifying the
    cruise segment distance and payload weight of the aicraft .

        Sources:
        N/A

        Assumptions:
        None 

        Inputs:
            vehicle             data structure for aircraft                  [-]
            mission             data structure for mission                   [-] 
            cruise_segment_tag  string of cruise segment                     [string]
            reserves            reserve fuel                                 [unitless] 
            
        Outputs: 
            payload_range       data structure of payload range properties   [m/s]
    """ 
    mass = vehicle.mass_properties
    if not mass.operating_empty:
        print("Error calculating Payload Range Diagram: vehicle Operating Empty Weight is undefined.")
        return True
    else:
        OEW = mass.operating_empty

    if not mass.max_payload:
        print("Error calculating Payload Range Diagram: vehicle Maximum Payload Weight is undefined.")
        return True
    else:
        MaxPLD = mass.max_payload

    if not mass.max_takeoff:
        print("Error calculating Payload Range Diagram: vehicle Maximum Payload Weight is undefined.")
        return True
    else:
        MTOW = mass.max_takeoff

    # Define Diagram Points
    # Point = [Value at Maximum Payload Range,  Value at Ferry Range]
    TOW =   [MTOW,      OEW]    # Takeoff Weights
    PLD =   [MaxPLD,    0.]     # Payload Weights

    # Initialize Range Array
    R = np.zeros(2)

    # Calculate Vehicle Range for Max Payload and Ferry Conditions
    for i in range(2):
        mission.segments[0].analyses.weights.vehicle.mass_properties.takeoff = TOW[i]
        results = mission.evaluate()
        segment = results.segments[cruise_segment_tag]
        R[i]    = segment.conditions.frames.inertial.position_vector[-1,0] 

    # Insert Starting Point for Diagram Construction
    R   = np.insert(R, 0, 0)
    PLD = np.insert(PLD, 0, MaxPLD)
    TOW = np.insert(TOW, 0, 0)

    # Pack Results
    payload_range = Data()
    payload_range.range             = np.array(R)
    payload_range.payload           = np.array(PLD)
    payload_range.takeoff_weight    = np.array(TOW)

    if plot_diagram: 
        # get plotting style 
        ps      = plot_style()  
    
        parameters = {'axes.labelsize': ps.axis_font_size,
                      'xtick.labelsize': ps.axis_font_size,
                      'ytick.labelsize': ps.axis_font_size,
                      'axes.titlesize': ps.title_font_size}
        plt.rcParams.update(parameters)

        fig  = plt.figure('Electric_Payload_Range_Diagram')
        axis = fig.add_subplot(1,1,1)        
        axis.plot(payload_range.range /Units.nmi, payload_range.payload,color = 'k', linewidth = ps.line_width )
        axis.set_xlabel('Range (nautical miles)')
        axis.set_ylabel('Payload (kg)')
        axis.set_title('Payload Range Diagram')
        set_axes(axis) 
        fig.tight_layout()

    return payload_range

# ---------------------------------------------------------------------
#   Define the Configurations
# ---------------------------------------------------------------------

def configs_setup(vehicle):
 
    
    # ------------------------------------------------------------------
    #   Initialize Configurations
    # ------------------------------------------------------------------

    configs     = RCAIDE.Library.Components.Configs.Config.Container() 
    base_config = RCAIDE.Library.Components.Configs.Config(vehicle)
    base_config.tag = 'base'  
    configs.append(base_config)
 
    return configs
  
def analyses_setup(configs,weights,aerodynamics):

    analyses = RCAIDE.Framework.Analyses.Analysis.Container()

    # build a base analysis for each config
    for tag,config in configs.items():
        analysis = base_analysis(config,weights,aerodynamics)
        analyses[tag] = analysis

    return analyses

def base_analysis(vehicle,weights,aerodynamics):

    # ------------------------------------------------------------------
    #   Initialize the Analyses
    # ------------------------------------------------------------------     
    analyses = RCAIDE.Framework.Analyses.Vehicle() 
    
    # ------------------------------------------------------------------
    #  Weights 
    analyses.append(weights)

    # ------------------------------------------------------------------
    #  Aerodynamics Analysis    
    analyses.append(aerodynamics)   

    # ------------------------------------------------------------------
    #  Energy
    energy          = RCAIDE.Framework.Analyses.Energy.Energy()
    energy.vehicle  = vehicle 
    analyses.append(energy)

    # ------------------------------------------------------------------
    #  Planet Analysis
    planet = RCAIDE.Framework.Analyses.Planets.Earth()
    analyses.append(planet)

    # ------------------------------------------------------------------
    #  Atmosphere Analysis
    atmosphere = RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976()
    atmosphere.features.planet = planet.features
    analyses.append(atmosphere)   

    # done!
    return analyses    

# ----------------------------------------------------------------------
#   Define the Mission
# ----------------------------------------------------------------------
 
def mission_setup(analyses,electric_aircraft,assigned_propulsors, cruise_airspeed, cruise_altitude, max_range_guess): 
    
    # ------------------------------------------------------------------
    #   Initialize the Mission
    # ------------------------------------------------------------------

    mission = RCAIDE.Framework.Mission.Sequential_Segments()
    mission.tag = 'mission'
  
    Segments = RCAIDE.Framework.Mission.Segments 
    base_segment = Segments.Segment()
    base_segment.state.numerics.number_of_control_points  = 3   

    # ------------------------------------------------------------------    
    #   Cruise Segment 
    # ------------------------------------------------------------------    

    segment = Segments.Cruise.Constant_Speed_Constant_Altitude(base_segment)
    segment.tag = "cruise" 
    segment.analyses.extend( analyses.base )
    if electric_aircraft: 
        segment.initial_battery_state_of_charge               = 1.0
     
    segment.altitude  = cruise_altitude
    segment.air_speed = cruise_airspeed
    segment.distance  = max_range_guess
    
    # define flight dynamics to model 
    segment.flight_dynamics.force_x                      = True  
    segment.flight_dynamics.force_z                      = True     
    
    # define flight controls 
    segment.assigned_control_variables.throttle.active               = True           
    segment.assigned_control_variables.throttle.assigned_propulsors  = assigned_propulsors
    segment.assigned_control_variables.body_angle.active             = True                
    
    mission.append_segment(segment) 
 

    return mission

def missions_setup(mission): 
 
    missions     = RCAIDE.Framework.Mission.Missions() 
    mission.tag  = 'base_mission'
    missions.append(mission)
 
    return missions   




# VLM.py
# 
# Created:  Oct 2020, E. Botero
# Modified: May 2021, E. Botero   
#           Jul 2021, A. Blaufox     

# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------

# package imports 
import numpy as np 
from RCAIDE.Framework.Core import Data
from .compute_wing_induced_velocity      import compute_wing_induced_velocity
from .generate_vortex_distribution       import generate_vortex_distribution 
from .compute_RHS_matrix                 import compute_RHS_matrix 

# ----------------------------------------------------------------------
#  Vortex Lattice
# ----------------------------------------------------------------------

def VLM(conditions,settings,geometry):
    """Uses the vortex lattice method to compute the lift, induced drag and moment coefficients.
    
    The user has the option to discretize control surfaces using the boolean settings.discretize_control_surfaces.
    The user should be forwarned that this will cause very slight differences in results for 0 deflection due to
    the slightly different discretization.
    
    The user has the option to use the boundary conditions and induced velocities from either RCAIDE
    or VORLAX. See build_RHS in compute_RHS_matrix.py for more details.
    
    By default in Vortex_Lattice, VLM performs calculations based on panel coordinates with float32 precision. 
    The user may also choose to use float16 or float64, but be warned that the latter can be memory intensive.
    
    The user should note that fully capitalized variables correspond to a VORLAX variable of the same name
    
    
    Assumptions:
    The user provides either global discretezation (number_spanwise/chordwise_vortices) or
    separate discretization (wing/fuselage_spanwise/chordwise_vortices) in settings, not both.
    The set of settings not being used should be set to None.
    
    The VLM requires that the user provide a non-zero velocity that matches mach number. For
    surrogate training cases at mach 0, VLM uses a velocity of 1e-6 m/s

    
    Source:
    1. Miranda, Luis R., Robert D. Elliot, and William M. Baker. "A generalized vortex 
    lattice method for subsonic and supersonic flow applications." (1977). (NASA CR)
    
    2. VORLAX Source Code

    
    Inputs:
    geometry.
       reference_area                          [m^2]
       wing.
         spans.projected                       [m]
         chords.root                           [m]
         chords.tip                            [m]
         sweeps.quarter_chord                  [radians]
         taper                                 [Unitless]
         twists.root                           [radians]
         twists.tip                            [radians]
         symmetric                             [Boolean]
         aspect_ratio                          [Unitless]
         areas.reference                       [m^2]
         vertical                              [Boolean]
         origin                                [m]
       fuselage.
        origin                                 [m]
        width                                  [m]
        heights.maximum                        [m]      
        lengths.nose                           [m]    
        lengths.tail                           [m]     
        lengths.total                          [m]     
        lengths.cabin                          [m]     
        fineness.nose                          [Unitless]
        fineness.tail                          [Unitless]
        
    settings.number_of_spanwise_vortices       [Unitless]  <---|
    settings.number_of_chordwise_vortices      [Unitless]  <---|
                                                               |--Either/or; see generate_vortex_distribution() for more details
    settings.wing_spanwise_vortices            [Unitless]  <---|
    settings.wing_chordwise_vortices           [Unitless]  <---|
    settings.fuselage_spanwise_vortices        [Unitless]  <---|
    settings.fuselage_chordwise_vortices       [Unitless]  <---|  
       
    settings.use_surrogate                     [Unitless]
    settings.propeller_wake_model              [Unitless]
    settings.discretize_control_surfaces       [Boolean], set to True to generate control surface panels
    settings.use_VORLAX_matrix_calculation     [boolean]
    settings.floating_point_precision          [float16/32/64]
       
    conditions.aerodynamics.angles.alpha    [radians]
    conditions.aerodynamics.angles.beta   [radians]
    conditions.freestream.mach_number          [Unitless]
    conditions.freestream.velocity             [m/s]
    conditions.static_stability.pitch_rate    [radians/s]
    conditions.static_stability.roll_rate     [radians/s]
    conditions.static_stability.yaw_rate      [radians/s]
       
    
    Outputs:    
    results.
        CL                                     [Unitless], CLTOT in VORLAX
        CDi                                    [Unitless], CDTOT in VORLAX
        CM                                     [Unitless], CMTOT in VORLAX
        CY                                  [Unitless], Total y force coeff
        CRTOT                                  [Unitless], Rolling moment coeff (unscaled)
        CL_mom                                 [Unitless], Rolling moment coeff (scaled by b_ref)
        CNTOT                                  [Unitless], Yawing  moment coeff (unscaled)
        CN                                 [Unitless], Yawing  moment coeff (scaled by b_ref)
        CL_wing                                [Unitless], CL  of each wing
        CDi_wing                               [Unitless], CDi of each wing
        cl_y                                   [Unitless], CL  of each strip
        cdi_y                                  [Unitless], CDi of each strip
        alpha_i                                [radians] , Induced angle of each strip in each wing (array of numpy arrays)
        CP                                     [Unitless], Pressure coefficient of each panel
        gamma                                  [Unitless], Vortex strengths of each panel

    
    Properties Used:
    N/A
    """ 
    # unpack settings----------------------------------------------------------------
    pwm        = settings.propeller_wake_model
    K_SPC      = settings.leading_edge_suction_multiplier
    Sref       = geometry.reference_area              

    # unpack geometry----------------------------------------------------------------
    # define point about which moment coefficient is computed
    if 'main_wing' in geometry.wings:
        c_bar      = geometry.wings['main_wing'].chords.mean_aerodynamic
        x_mac      = geometry.wings['main_wing'].aerodynamic_center[0] + geometry.wings['main_wing'].origin[0][0]
        z_mac      = geometry.wings['main_wing'].aerodynamic_center[2] + geometry.wings['main_wing'].origin[0][2]
        b_ref      = geometry.wings['main_wing'].spans.projected
    else:
        c_bar  = 0.
        x_mac  = 0.
        b_ref = 0.
        for wing in geometry.wings:
            if wing.vertical == False:
                if c_bar <= wing.chords.mean_aerodynamic:
                    c_bar  = wing.chords.mean_aerodynamic
                    x_mac  = wing.aerodynamic_center[0] + wing.origin[0][0]
                    z_mac  = wing.aerodynamic_center[2] + wing.origin[0][2]
                    b_ref  = wing.spans.projected

    x_cg       = geometry.mass_properties.center_of_gravity[0][0]
    z_cg       = geometry.mass_properties.center_of_gravity[0][2]
    if x_cg == 0.0:
        x_m = x_mac 
        z_m = z_mac
    else:
        x_m = x_cg
        z_m = z_cg
        
    # unpack conditions--------------------------------------------------------------
    aoa  = conditions.aerodynamics.angles.alpha   # angle of attack  
    mach = conditions.freestream.mach_number         # mach number
    ones = np.atleast_2d(np.ones_like(mach)) 
    len_mach = len(mach)
    
    #For angular values, VORLAX uses degrees by default to radians via DTR (degrees to rads). 
    #RCAIDE uses radians and its Units system. All algular variables will be in radians or var*Units.degrees
    PSI       = conditions.aerodynamics.angles.beta    
    PITCHQ    = conditions.static_stability.pitch_rate              
    ROLLQ     = conditions.static_stability.roll_rate             
    YAWQ      = conditions.static_stability.yaw_rate 
    VINF      = conditions.freestream.velocity    
       
    #freestream 0 velocity safeguard
    if not conditions.freestream.velocity.all():
        if settings.use_surrogate:
            velocity                       = conditions.freestream.velocity
            velocity[velocity==0]          = np.ones(len(velocity[velocity==0])) * 1e-6
            conditions.freestream.velocity = velocity
        else:
            raise AssertionError("VLM requires that conditions.freestream.velocity be specified and non-zero")    

    # ---------------------------------------------------------------------------------------
    # STEPS 1-9: Generate Panelization and Vortex Distribution
    # ------------------ --------------------------------------------------------------------    
    # generate vortex distribution (VLM steps 1-9)
    VD   = generate_vortex_distribution(geometry,settings)  
    
    if not VD.is_postprocessed:
        raise ValueError('postprocess_VD has not been called since the panels have been modified')
    
    # Unpack vortex distribution
    n_cp         = VD.n_cp 
    n_sw         = VD.n_sw
    CHORD        = VD.chord_lengths
    chord_breaks = VD.chordwise_breaks
    span_breaks  = VD.spanwise_breaks
    RNMAX        = VD.panels_per_strip    
    LE_ind       = VD.leading_edge_indices
    ZETA         = VD.tangent_incidence_angle
    RK           = VD.chordwise_panel_number
    
    exposed_leading_edge_flag = VD.exposed_leading_edge_flag
    
    YAH = VD.YAH*1.  
    YBH = VD.YBH*1.
    
    XA1 = VD.XA1*1.
    XB1 = VD.XB1*1.
    YA1 = VD.YA1
    YB1 = VD.YB1    
    ZA1 = VD.ZA1
    ZB1 = VD.ZB1  
    
    XCH = VD.XCH
    
    XA_TE =  VD.XA_TE
    XB_TE =  VD.XB_TE
    YA_TE =  VD.YA_TE
    YB_TE =  VD.YB_TE
    ZA_TE =  VD.ZA_TE
    ZB_TE =  VD.ZB_TE     
     
    SLOPE = VD.SLOPE
    SLE   = VD.SLE
    D     = VD.D
    
    # Compute X and Z BAR ouside of generate_vortex_distribution to avoid requiring x_m and z_m as inputs
    XBAR    = np.ones(sum(LE_ind)) * x_m
    ZBAR    = np.ones(sum(LE_ind)) * z_m
    VD.XBAR = XBAR
    VD.ZBAR = ZBAR
    
    # ---------------------------------------------------------------------------------------
    # STEP 10: Generate A and RHS matrices from VD and geometry
    # ------------------ --------------------------------------------------------------------    
    # Compute flow tangency conditions
    phi   = np.arctan((VD.ZBC - VD.ZAC)/(VD.YBC - VD.YAC))*ones # dihedral angle 
    delta = np.arctan((VD.ZC - VD.ZCH)/((VD.XC - VD.XCH)*ones)) # mean camber surface angle 

    # Build the RHS vector    
    rhs = compute_RHS_matrix(delta,phi,conditions,settings,geometry,pwm) 
    RHS     = rhs.RHS*1
    ONSET   = rhs.ONSET*1

    # Build induced velocity matrix, C_mn
    # This is not affected by AoA, so we can use unique mach numbers only
    m_unique, inv = np.unique(mach,return_inverse=True)
    m_unique      = np.atleast_2d(m_unique).T
    C_mn_small, s, RFLAG_small, EW_small = compute_wing_induced_velocity(VD,m_unique,compute_EW=True)
    
    C_mn  = C_mn_small[inv,:,:,:]
    RFLAG = RFLAG_small[inv,:]
    EW    = EW_small[inv,:,:]

    # Turn off sonic vortices when Mach>1
    RHS = RHS*RFLAG
    
    # Build Aerodynamic Influence Coefficient Matrix
    use_VORLAX_induced_velocity = settings.use_VORLAX_matrix_calculation
    if not use_VORLAX_induced_velocity:
        A =   np.multiply(C_mn[:,:,:,0],np.atleast_3d(np.sin(delta)*np.cos(phi))) \
            + np.multiply(C_mn[:,:,:,1],np.atleast_3d(np.cos(delta)*np.sin(phi))) \
            - np.multiply(C_mn[:,:,:,2],np.atleast_3d(np.cos(phi)*np.cos(delta)))   # validated from book eqn 7.42 
    else:
        A = EW

    # Compute vortex strength
    GAMMA  = np.linalg.solve(A,RHS)

    # ---------------------------------------------------------------------------------------
    # STEP 11: Compute Pressure Coefficient
    # ------------------ --------------------------------------------------------------------   
    #VORLAX subroutine = PRESS
                  
    # spanwise strip exposure flag, always 0 for RCAIDE's infinitely thin airfoils. Needs to change if thick airfoils added
    RJTS = 0                         
    
    # COMPUTE FREE-STREAM AND ONSET FLOW PARAMETERS. Used throughout the remainder of VLM
    B2     = np.tile((mach**2 - 1),n_cp)
    SINALF = np.sin(aoa)
    COSALF = np.cos(aoa)
    TANALF = np.tan(aoa)
    SINPSI = np.sin(PSI)
    COPSI  = np.cos(PSI)
    COSIN  = COSALF *SINPSI *2.0
    COSINP = COSALF *SINPSI
    COSCOS = COSALF *COPSI
    PITCH  = PITCHQ /VINF
    ROLL   = ROLLQ /VINF
    YAW    = YAWQ /VINF    
    
    # reshape CHORD
    CHORD  = CHORD[0,:]
    CHORD_strip = CHORD[LE_ind]

    # COMPUTE EFFECT OF SIDESLIP on DCP intermediate variables. needs change if cosine chorwise spacing added
    FORAXL = COSCOS
    FORLAT = COSIN
    
    TAN_LE = (XB1[LE_ind] - XA1[LE_ind])/ \
                np.sqrt((ZB1[LE_ind]-ZA1[LE_ind])**2 + \
                        (YB1[LE_ind]-YA1[LE_ind])**2)  
    TAN_TE = (XB_TE - XA_TE)/ np.sqrt((ZB_TE-ZA_TE)**2 + (YB_TE-YA_TE)**2) # _TE variables already have np.repeat built in 
    TAN_LE = np.broadcast_to(np.repeat(TAN_LE,RNMAX[LE_ind]),np.shape(B2)) 
    TAN_TE = np.broadcast_to(TAN_TE                         ,np.shape(B2))    
    
    TNL    = TAN_LE * 1 # VORLAX's SIGN variable not needed, as these are taken directly from geometry
    TNT    = TAN_TE * 1
    XIA    = np.broadcast_to((RK-1)/RNMAX, np.shape(B2))
    XIB    = np.broadcast_to((RK  )/RNMAX, np.shape(B2))
    TANA   = TNL *(1. - XIA) + TNT *XIA
    TANB   = TNL *(1. - XIB) + TNT *XIB
    
    # cumsum GANT loop if KTOP > 0 (don't actually need KTOP with vectorized arrays and np.roll)
    GFX    = np.tile((1 /CHORD), (len_mach,1))
    GANT   = strip_cumsum(GFX*GAMMA, chord_breaks, RNMAX[LE_ind])
    GANT   = np.roll(GANT,1)
    GANT[:,LE_ind]   = 0 
    
    GLAT   = GANT *(TANA - TANB) - GFX *GAMMA *TANB
    COS_DL = (YBH-YAH)[LE_ind]/D
    cos_DL = np.broadcast_to(np.repeat(COS_DL,RNMAX[LE_ind]),np.shape(B2))
    DCPSID = FORLAT * cos_DL *GLAT /(XIB - XIA)
    FACTOR = FORAXL + ONSET
    
    # COMPUTE LOAD COEFFICIENT
    GNET = GAMMA*FACTOR
    GNET = GNET *RNMAX /CHORD
    DCP  = 2*GNET + DCPSID
    CP   = DCP

    # ---------------------------------------------------------------------------------------
    # STEP 12: Compute aerodynamic coefficients 
    # ------------------ -------------------------------------------------------------------- 
    #VORLAX subroutine = AERO

    # Work panel by panel
    SURF = np.array(VD.wing_areas)
    SREF = Sref  

    # Flip coordinates on the other side of the wing
    boolean = YBH<0. 
    XA1[boolean], XB1[boolean] = XB1[boolean], XA1[boolean]
    YAH[boolean], YBH[boolean] = YBH[boolean], YAH[boolean]

    # Leading edge sweep. VORLAX does it panel by panel. This will be spanwise.
    TLE   = TAN_LE[:,LE_ind]
    B2_LE = B2[:,LE_ind]
    T2    = TLE*TLE
    STB   = np.zeros_like(B2_LE)
    STB[B2_LE<T2] = np.sqrt(T2[B2_LE<T2]-B2_LE[B2_LE<T2])
    
    # DL IS THE DIHEDRAL ANGLE (WITH RESPECT TO THE X-Y PLANE) OF
    # THE IR STREAMWISE STRIP OF HORSESHOE VORTICES. 
    COD = np.cos(phi[0,LE_ind])  # Just the LE values
    
    SID = np.sin(phi[0,LE_ind])  # Just the LE values

    # Now on to each strip
    PION = 2.0 /RNMAX
    ADC  = 0.5*PION

    # XLE = LOCATION OF FIRST VORTEX MIDPOINT IN FRACTION OF CHORD.
    XLE = 0.125 *PION
    
    GAF = 0.5 + 0.5 *RJTS**2

    # CORMED IS LENGTH OF STRIP CENTERLINE BETWEEN LOAD POINT
    # AND TRAILING EDGE THIS PARAMETER IS USED IN THE COMPUTATION
    # OF THE STRIP ROLLING COUPLE CONTRIBUTION DUE TO SIDESLIP.
    X      = XCH                       #x-coord of load point (horseshoe centroid)
    XTE    = (VD.XA_TE + VD.XB_TE)/2   #Trailing edge x-coord behind the control point  
    CORMED = XTE - X   

    # SINF REFERENCES THE LOAD CONTRIBUTION OF IRT-VORTEX TO THE
    # STRIP NOMINAL AREA, I.E., AREA OF STRIP ASSUMING CONSTANT
    # (CHORDWISE) HORSESHOE SPAN.    
    SINF = ADC * DCP # The horshoe span lengths have been removed since VST/VSS == 1 always

    # Split into chordwise strengths and sum into strips    
    # SICPLE = COUPLE (ABOUT STRIP CENTERLINE) DUE TO SIDESLIP.
    CNC    = np.add.reduceat(SINF       ,chord_breaks,axis=1)
    SICPLE = np.add.reduceat(SINF*CORMED,chord_breaks,axis=1)

    # COMPUTE SLOPE (TX) WITH RESPECT TO X-AXIS AT LOAD POINTS BY INTER
    # POLATING BETWEEN CONTROL POINTS AND TAKING INTO ACCOUNT THE LOCAL
    # INCIDENCE.    
    XX   = (RK - .75) *PION /2.0
    TX    = SLOPE - ZETA
    CAXL  = -SINF*TX/(1.0+TX**2) # These are the axial forces on each panel
    BMLE  = (XLE-XX)*SINF        # These are moment on each panel
    
    # Sum onto the panel
    CAXL = np.add.reduceat(CAXL,chord_breaks,axis=1)
    BMLE = np.add.reduceat(BMLE,chord_breaks,axis=1)
    
    SICPLE *= (-1) * COSIN * COD * GAF
    DCP_LE = DCP[:,LE_ind]
    
    # COMPUTE LEADING EDGE THRUST COEFF. (CSUC) BY CALCULATING
    # THE TOTAL INDUCED FLOW AT THE LEADING EDGE. THIS COMPUTATION
    # ONLY PERFORMED FOR COSINE CHORDWISE SPACING (LAX = 0).    
    # ** TO DO ** Add cosine spacing (earlier in VLM) to properly capture the magnitude of these earlier.
    # Right now, this computation still happens with linear spacing, though its effects are underestimated.
    CLE = compute_rotation_effects(VD, settings, EW, GAMMA, len_mach, X, CHORD, XLE, XBAR, 
                                   rhs, COSINP, SINALF,COSCOS, PITCH, ROLL, YAW, STB, RNMAX)    
    
    # Leading edge suction multiplier. See documentation. This is a negative integer if used
    # Default to 1 unless specified otherwise
    SPC  = K_SPC*np.ones_like(DCP_LE)
    
    # If the vehicle is subsonic and there is vortex lift enabled then SPC changes to -1
    VL   = np.repeat(VD.vortex_lift,n_sw)
    m_b  = np.atleast_2d(mach[:,0]<1.)
    SPC_cond      = VL*m_b.T
    SPC[SPC_cond] = -1.
    SPC           = SPC * exposed_leading_edge_flag
    
    CLE  = CLE + 0.5* DCP_LE *np.sqrt(XLE[LE_ind])
    CSUC = 0.5*np.pi*np.abs(SPC)*(CLE**2)*STB 

    # TFX AND TFZ ARE THE COMPONENTS OF LEADING EDGE FORCE VECTOR ALONG
    # ALONG THE X AND Z BODY AXES.   
    
    SLE  = SLOPE[LE_ind]
    ZETA = ZETA[LE_ind]
    XCOS = np.broadcast_to(np.cos(SLE-ZETA),np.shape(DCP_LE))
    XSIN = np.broadcast_to(np.sin(SLE-ZETA),np.shape(DCP_LE))
    TFX  =  1.*XCOS
    TFZ  = -1.*XSIN

    # If a negative number is used for SPC a different correction is used. See VORLAX documentation for Lan reference
    TFX[SPC<0] = XSIN[SPC<0]*np.sign(DCP_LE)[SPC<0]
    TFZ[SPC<0] = np.abs(XCOS)[SPC<0]*np.sign(DCP_LE)[SPC<0]

    CAXL = CAXL - TFX*CSUC
    
    # Add a dimension into the suction to be chordwise
    CNC   = CNC + CSUC*np.sqrt(1+T2)*TFZ
    
    # FCOS AND FSIN ARE THE COSINE AND SINE OF THE ANGLE BETWEEN
    # THE CHORDLINE OF THE IR-STRIP AND THE X-AXIS    
    FCOS = np.cos(ZETA)
    FSIN = np.sin(ZETA)
    
    # BFX, BFY, AND BFZ ARE THE COMPONENTS ALONG THE BODY AXES
    # OF THE STRIP FORCE CONTRIBUTION.
    BFX = -  CNC *FSIN + CAXL *FCOS
    BFY = - (CNC *FCOS + CAXL *FSIN) *SID
    BFZ =   (CNC *FCOS + CAXL *FSIN) *COD

    # CONVERT CNC FROM CN INTO CNC (COEFF. *CHORD).
    CNC  = CNC  * CHORD_strip
    BMLE = BMLE * CHORD_strip

    # BMX, BMY, AND BMZ ARE THE COMPONENTS ALONG THE BODY AXES
    # OF THE STRIP MOMENT (ABOUT MOM. REF. POINT) CONTRIBUTION.
    X      = VD.XCH[LE_ind]  # These are all LE values
    Y      = VD.YCH[LE_ind]  # These are all LE values
    Z      = VD.ZCH[LE_ind]  # These are all LE values
    BMX    = BFZ * Y - BFY * (Z - ZBAR)
    BMX    = BMX + SICPLE
    BMY    = BMLE * COD + BFX * (Z - ZBAR) - BFZ * (X - XBAR)
    BMZ    = BMLE * SID - BFX * Y + BFY * (X - XBAR)
    CDC    = BFZ * SINALF +  (BFX *COPSI + BFY *SINPSI) * COSALF
    CDC    = CDC * CHORD_strip 

    ES     = 2*s[0,LE_ind]
    STRIP  = ES *CHORD_strip
    LIFT   = (BFZ *COSALF - (BFX *COPSI + BFY *SINPSI) *SINALF)*STRIP   
    DRAG   = CDC*ES 
    MOMENT = STRIP * (BMY *COPSI - BMX *SINPSI)  
    FY     = (BFY *COPSI - BFX *SINPSI) *STRIP
    RM     = STRIP *(BMX *COSALF *COPSI + BMY *COSALF *SINPSI + BMZ *SINALF)
    YM     = STRIP *(BMZ *COSALF - (BMX *COPSI + BMY *SINPSI) *SINALF)

    # Now calculate the coefficients for each wing
    cl_y     = LIFT/CHORD_strip/ES
    cdi_y    = DRAG/CHORD_strip/ES
    CL_wing  = np.add.reduceat(LIFT,span_breaks,axis=1)/SURF
    CDi_wing = np.add.reduceat(DRAG,span_breaks,axis=1)/SURF
    alpha_i  = np.hsplit(np.arctan(cdi_y/cl_y),span_breaks[1:])
    
    # Now calculate total coefficients
    CL       = np.atleast_2d(np.sum(LIFT,axis=1)/SREF).T          # CLTOT in VORLAX
    CDi      = np.atleast_2d(np.sum(DRAG,axis=1)/SREF).T          # CDTOT in VORLAX
    CX       =  ( TANALF * CL - CDi)/(COSALF - SINALF*TANALF)
    CZ       = (CDi+ CX*COSALF)/SINALF 
    CM       = np.atleast_2d(np.sum(MOMENT,axis=1)/SREF).T/c_bar  # CMTOT in VORLAX 1. check deflection is accounted for correctly. 2. check this is right
    CY       = np.atleast_2d(np.sum(FY,axis=1)/SREF).T   # total y force coeff
    CRTOT    = np.atleast_2d(np.sum(RM,axis=1)/SREF).T   # rolling moment coeff (unscaled)
    CL_mom   = CRTOT/b_ref*(-1)                         # rolling moment coeff
    CNTOT    = np.atleast_2d(np.sum(YM,axis=1)/SREF).T   # yawing  moment coeff (unscaled)
    CN       = CNTOT/b_ref*(-1)                         # yawing  moment coeff

    # ---------------------------------------------------------------------------------------
    # STEP 13: Pack outputs
    # ------------------ --------------------------------------------------------------------     
    precision      = settings.floating_point_precision
    
    #VORLAX _TOT outputs
    results = Data()
    # force coefficients
    results.S_ref      = Sref
    results.b_ref      = b_ref
    results.c_ref      = c_bar  
    results.X_ref      = x_m
    results.Y_ref      = 0
    results.Z_ref      = z_m
    
    results.CL         =  CL
    results.CDi        =  CDi
    
    results.CX         =  CX
    results.CY         =  CY 
    results.CZ         = -CZ
    
    results.CL_mom     =  CL_mom 
    results.CM         =  CM  
    results.CN         =  CN
    
    #other RCAIDE outputs
    results.CL_wing        = CL_wing   
    results.CDi_wing       = CDi_wing 
    results.cl_y           = cl_y   
    results.cdi_y          = cdi_y       
    results.alpha_i        = alpha_i  
    results.CP             = np.array(CP    , dtype=precision)
    results.gamma          = np.array(GAMMA , dtype=precision)
    results.VD             = VD
    results.V_distribution = rhs.V_distribution
    results.V_x            = rhs.Vx_ind_total
    results.V_z            = rhs.Vz_ind_total
    
    return results

# ----------------------------------------------------------------------
#  CLE rotation effects helper function
# ----------------------------------------------------------------------
def compute_rotation_effects(VD, settings, EW_small, GAMMA, len_mach, X, CHORD, XLE, XBAR, 
                             rhs, COSINP, SINALF,COSCOS, PITCH, ROLL, YAW, STB, RNMAX):
    """ This computes the effects of the freestream and aircraft rotation rate on 
    CLE, the induced flow at the leading edge
    
    Assumptions:
    Several of the values needed in this calculation have been computed earlier and stored in VD
    
    Normally, VORLAX skips the calculation implemented in this function for linear 
    chordwise spacing (the if statement below). However, since the trends are correct, 
    albeit underestimated, this calculation is being forced here.    
    """
    LE_ind      = VD.leading_edge_indices
    RNMAX       = VD.panels_per_strip

    ##spacing = settings.spanwise_cosine_spacing
    ##if spacing == False: # linear spacing is LAX==1 in VORLAX
    ##    return 0 #CLE not calculated till later for linear spacing
    
    # Computate rotational effects (pitch, roll, yaw rates) on LE suction
    # pick leading edge strip values for EW and reshape GAMMA -> gamma accordingly
    EW    = EW_small[: ,LE_ind, :]
    n_tot_strips = EW.shape[1]
    gamma = np.array(np.split(np.repeat(GAMMA, n_tot_strips, axis=0), len_mach))
    CLE = (EW*gamma).sum(axis=2)
    
    # Up till EFFINC, some of the following values were computed in compute_RHS_matrix().
    #     EFFINC and ALOC are calculated the exact same way, except for the XGIRO term.
    # LOCATE VORTEX LATTICE CONTROL POINT WITH RESPECT TO THE
    # ROTATION CENTER (XBAR, 0, ZBAR). THE RELATIVE COORDINATES
    # ARE XGIRO, YGIRO, AND ZGIRO. 
    XGIRO = X - CHORD*XLE - np.repeat(XBAR, RNMAX[LE_ind])
    YGIRO = rhs.YGIRO
    ZGIRO = rhs.ZGIRO
    
    # VX, VY, VZ ARE THE FLOW ONSET VELOCITY COMPONENTS AT THE LEADING
    # EDGE (STRIP MIDPOINT). VX, VY, VZ AND THE ROTATION RATES ARE
    # REFERENCED TO THE FREE STREAM VELOCITY.     
    VX = (COSCOS - PITCH*ZGIRO + YAW  *YGIRO)  
    VY = (COSINP - YAW  *XGIRO + ROLL *ZGIRO)  
    VZ = (SINALF - ROLL *YGIRO + PITCH*XGIRO)

    # CCNTL, SCNTL, SID, and COD were computed in compute_RHS_matrix()
    
    # EFFINC = COMPONENT OF ONSET FLOW ALONG NORMAL TO CAMBERLINE AT
    #          LEADING EDGE.
    EFFINC = VX *rhs.SCNTL + VY *rhs.CCNTL *rhs.SID - VZ *rhs.CCNTL *rhs.COD 
    CLE = CLE - EFFINC[:,LE_ind] 
    CLE = np.where(STB > 0, CLE /RNMAX[LE_ind] /STB, CLE)
    
    return CLE

# ----------------------------------------------------------------------
#  Vectorized cumsum from indices
# ----------------------------------------------------------------------
def strip_cumsum(arr, chord_breaks, strip_lengths):
    """ Uses numpy to to compute a cumsum that resets along
    the leading edge of every strip.
    
    Assumptions:
    chordwise_breaks always starts at 0
    """    
    cumsum  = np.cumsum(arr, axis=1)
    offsets = cumsum[:,chord_breaks-1]
    offsets[:,0]  = 0
    offsets = np.repeat(offsets, strip_lengths, axis=1)
    return cumsum - offsets



def skip(*args,**kwarg):
    """This method can be used to replace default functions when
    no action is desired instead.

    Assumptions:
    N/A

    Source:
    N/A

    Inputs:
    None

    Outputs:
    None

    Properties Used:
    N/A
    """          
    pass



# RCAIDE/Methods/Aerodynamics/Fidelity_Zero/__init__.py
# 

""" RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
from .build_VLM_surrogates                    import build_VLM_surrogates  
from .compute_RHS_matrix                      import compute_RHS_matrix 
from .compute_wing_induced_velocity           import compute_wing_induced_velocity
from .deflect_control_surface                 import deflect_control_surfaces
from .extract_wing_collocation_points         import extract_wing_collocation_points
from .generate_VD_helpers                     import postprocess_VD, compute_panel_area, compute_unit_normal 
from .generate_vortex_distribution            import generate_vortex_distribution
from .make_VLM_wings                          import make_VLM_wings
from .train_VLM_surrogates                    import train_VLM_surrogates
from .VLM                                     import VLM
from .evaluate_VLM                            import *  




# RCAIDE/Library/Components/Propulsors/Converters/Compressor.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Feb 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
 # RCAIDE imports   
from RCAIDE.Library.Components                      import Component  
from RCAIDE.Library.Methods.Propulsors.Converters.Compressor.append_compressor_conditions import append_compressor_conditions

# ---------------------------------------------------------------------------------------------------------------------- 
#  Compressor 
# ---------------------------------------------------------------------------------------------------------------------- 
class Compressor(Component):
    """
    A compressor component model for gas turbine engines.

    Attributes
    ----------
    tag : str
        Identifier for the compressor. Default is 'Compressor'.
        
    polytropic_efficiency : float
        Efficiency of the compression process accounting for losses. Default is 1.0.
        
    pressure_ratio : float
        Ratio of outlet to inlet total pressure. Default is 1.0.

    Notes
    -----
    The Compressor class models the compression process in gas turbine engines.
    It calculates the work required to increase the pressure of the working fluid
    while accounting for losses through the polytropic efficiency.

    **Definitions**

    'Polytropic Efficiency'
        A measure of compression efficiency that accounts for the continuous
        nature of the compression process and real gas effects

    'Pressure Ratio'
        Ratio of exit to inlet total pressure, a key parameter determining
        the compressor's performance and work input required

    See Also
    --------
    RCAIDE.Library.Methods.Propulsors.Converters.Compressor
    """
    
    def __defaults__(self):
        """This sets the default values for the component to function.

        Assumptions:
            None

        Source:
            None 
        """          
        #set the default values
        self.tag                             = 'Compressor'
        self.polytropic_efficiency           = 1.0
        self.pressure_ratio                  = 1.0

    def append_operating_conditions(self,segment,propulsor): 
        propulsor_conditions =  segment.state.conditions.energy[propulsor.tag]
        append_compressor_conditions(self,segment,propulsor_conditions)
        return        


# RCAIDE/Library/Compoments/Thermal_Management/Reservoirs/Reservoir.py
# 
# Created:  Mar 2024, S. Shekar

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
from RCAIDE.Library.Components                                                      import Component 
from RCAIDE.Library.Attributes.Coolants.Glycol_Water                                import Glycol_Water
from RCAIDE.Library.Attributes.Materials.Polyetherimide                             import Polyetherimide
from RCAIDE.Library.Methods.Thermal_Management.Reservoirs.Reservoir_Tank            import compute_mixing_temperature, append_reservoir_conditions, append_reservoir_segment_conditions
from RCAIDE.Library.Plots.Thermal_Management.plot_reservoir_conditions              import plot_reservoir_conditions

# ----------------------------------------------------------------------
#  Reservoir
# ---------------------------------------------------------------------- 
class Reservoir(Component):
    """
    A class representing a coolant storage reservoir for thermal management systems.

    Attributes
    ----------
    tag : str
        Unique identifier for the reservoir, defaults to 'coolant_reservoir'
        
    material : Material
        Reservoir construction material, defaults to Polyetherimide()
        
    coolant : Coolant
        Fluid stored in reservoir, defaults to Glycol_Water()
        
    length : float
        Reservoir length in meters, defaults to 0.3
        
    width : float
        Reservoir width in meters, defaults to 0.3
        
    height : float
        Reservoir height in meters, defaults to 0.3
        
    thickness : float
        Wall thickness in meters, defaults to 5e-3
        
    surface_area : float
        Total surface area in square meters, calculated from dimensions
        
    volume : float
        Internal volume in cubic meters, calculated from dimensions
        
    mass_properties.mass : float
        Total mass including coolant, calculated from volume and coolant density

    Notes
    -----
    The reservoir serves as a thermal buffer and coolant storage system. It provides:
    
    * Coolant storage and thermal mass
    * Temperature stabilization through mixing
    * Volume compensation for thermal expansion
    * Air separation and pressure control

    **Definitions**

    'Mixing Temperature'
        Temperature resulting from combining flows of different temperatures
        
    'Surface Area'
        Total area for heat transfer with environment

    See Also
    --------
    RCAIDE.Library.Components.Thermal_Management.Heat_Exchangers.Cross_Flow_Heat_Exchanger
        Heat exchanger that interfaces with reservoir
    RCAIDE.Library.Components.Thermal_Management.Accessories.Pump
        Pump for circulating coolant from reservoir
    """

    def __defaults__(self):
        """
        Sets default values for the reservoir attributes.
        """
        self.tag              = 'coolant_reservoir'
        self.material         = Polyetherimide()
        self.coolant          = Glycol_Water()
        self.length           = 0.3
        self.width            = 0.3
        self.height           = 0.3
        self.thickness        = 5e-3
        self.surface_area     = 2*(self.length*self.width + self.width*self.height + 
                                  self.length*self.height)
        self.volume           = self.length*self.width*self.height
        self.mass_properties.mass = self.coolant.density*self.volume
        return
    
    def append_operating_conditions(self, segment, coolant_line):
        """
        Adds operating conditions for the reservoir to a mission segment.

        Parameters
        ----------
        segment : Data
            Mission segment to which conditions are being added
        coolant_line : Data
            Cooling system flow path information
        """
        append_reservoir_conditions(self, segment, coolant_line)
        return
    
    def append_segment_conditions(self, segment, coolant_line, conditions):
        """
        Adds specific segment conditions to the reservoir analysis.

        Parameters
        ----------
        segment : Data
            Mission segment being analyzed
        coolant_line : Data
            Cooling system flow path information
        conditions : Data
            Operating conditions for the segment
        """
        append_reservoir_segment_conditions(self, segment, coolant_line, conditions)
        return    

    def compute_reservior_coolant_temperature(self, state, coolant_line, delta_t, t_idx):
        """
        Calculates the mixed coolant temperature in the reservoir.

        Parameters
        ----------
        state : Data
            Current system state
        coolant_line : Data
            Cooling system flow path information
        delta_t : float
            Time step size
        t_idx : int
            Time index in the simulation
        """
        compute_mixing_temperature(self, state, coolant_line, delta_t, t_idx)
        return
    
    def plot_operating_conditions(self, results, coolant_line, save_filename, save_figure, 
                                show_legend, file_type, width, height):
        """
        Creates visualization plots of the reservoir operating conditions.

        Parameters
        ----------
        results : Data
            Simulation results data
        coolant_line : Data
            Cooling system flow path information
        save_filename : str
            Path for saving the plot
        save_figure : bool
            Flag to save the figure
        show_legend : bool
            Flag to display plot legend
        file_type : str
            Output file format
        width : float
            Plot width
        height : float
            Plot height
        """
        plot_reservoir_conditions(self, results, coolant_line, save_filename, 
                                save_figure, show_legend, file_type, width, height)
        return    



# RCAIDE/Library/Methods/Propulsor/Ducted_Fan_Propulsor/write_geometry.py
# 
# Created: Sep 2024, M. Clarke 

# ---------------------------------------------------------------------------------------------------------------------- 
#  Imports
# ----------------------------------------------------------------------------------------------------------------------
import RCAIDE
from RCAIDE.Framework.Core import Units 
from RCAIDE.Library.Methods.Geometry.Airfoil import  import_airfoil_geometry , compute_naca_4series
from .purge_files import purge_files   
import numpy as  np
 
# ---------------------------------------------------------------------------------------------------------------------- 
# Write Geometry 
# ---------------------------------------------------------------------------------------------------------------------- 
def write_geometry(dfdc_object,run_script_path):
    """This function writes the translated aircraft geometry into text file read 
    by DFDC when it is called 
    """     
    # unpack inputs
    case_file   = dfdc_object.geometry.tag + '.case' 
    purge_files([case_file]) 
    geometry             = open(case_file,'w') 
    with open(case_file,'w') as geometry:
        make_header_text(dfdc_object,geometry)
        make_hub_text(dfdc_object,geometry)
        make_separator_text(dfdc_object,geometry)  
        make_duct_text(dfdc_object,geometry)
    return

def make_header_text(dfdc_object,geometry):  
    """This function writes the header using the template required for the DFDC executable to read
 
    """      
    header_1_text_template = \
'''DFDC Version  0.70 
Bladed rotor + actdisk stator test case                                             

OPER
!        Vinf         Vref          RPM          RPM
  {0}   {1}       {2}          0.0    
!         Rho          Vso          Rmu           Alt
   1.2260       340.00      0.17800E-04  0.00000E+00
!       XDwake        Nwake
  0.80000               20
!       Lwkrlx
            F
ENDOPER

AERO
!  #sections
     1
!   Xisection
  0.00000E+00
!       A0deg        dCLdA        CLmax         CLmin
  0.00000E+00   6.2800       1.5000      -1.0000    
!  dCLdAstall     dCLstall      Cmconst         Mcrit
  0.50000      0.20000      0.00000E+00  0.70000    
!       CDmin      CLCDmin     dCDdCL^2
  0.12000E-01  0.10000      0.50000E-02
!       REref        REexp
  0.20000E+06  0.35000    
ENDAERO

ROTOR
!       Xdisk        Nblds       NRsta
  {3}                {4}           {5}
!  #stations
    {6}
!           r        Chord         Beta
'''
    
    length                = dfdc_object.geometry.length
    x_disc_rotor          = dfdc_object.geometry.rotor.percent_x_location * length
    number_of_blades      = dfdc_object.geometry.number_of_rotor_blades 
    number_of_end_points  = dfdc_object.geometry.number_of_radial_stations + 1 
    number_of_stations    = dfdc_object.geometry.number_of_radial_stations  
    hub_radius            = dfdc_object.geometry.hub_radius
    clearance             = dfdc_object.geometry.blade_clearance
    design_RPM            = dfdc_object.geometry.cruise.design_angular_velocity /Units.rpm 
    tip_radius            = dfdc_object.geometry.tip_radius  
    V_inf                 = dfdc_object.geometry.cruise.design_freestream_velocity  
    V_ref                 = dfdc_object.geometry.cruise.design_reference_velocity  
    header_1_text         = header_1_text_template.format(V_inf,V_ref, design_RPM,x_disc_rotor,number_of_blades,number_of_end_points, number_of_stations)
   
    geometry.write(header_1_text)
    
    station_chords = np.linspace(0.4,0.3, number_of_stations) *tip_radius
    station_twists = np.linspace(77, 30, number_of_stations)
    station_radii  = np.linspace(hub_radius+clearance, tip_radius-clearance,number_of_stations )
    for i in range(number_of_stations): 
        station_radius = station_radii[i]
        station_chord  = station_chords[i] 
        station_twist  = station_twists[i]
        case_text = '  ' + format(station_radius, '.6f')+ "  " + format(station_chord, '.6f')+ "   " + format(station_twist, '.6f') + "\n" 
        geometry.write(case_text)
        
    header_2_text_template = \
'''ENDROTOR


AERO
!  #sections
     1
!   Xisection
  0.00000E+00
!       A0deg        dCLdA        CLmax         CLmin
  0.00000E+00   6.2800       1.0000      -1.5000    
!  dCLdAstall     dCLstall      Cmconst         Mcrit
  0.50000      0.20000      0.00000E+00  0.70000    
!       CDmin      CLCDmin     dCDdCL^2
  0.12000E-01 -0.10000      0.50000E-02
!       REref        REexp          TOC      dCDdCL^2
  0.20000E+06  0.35000      0.10000      0.20000E-01
ENDAERO

ACTDISK
! Xdisk   NRPdef
  {0}     15
! #stations
  3
! r     BGam
 0.02   -10.0
 0.06   -10.0
 0.10   -10.0
ENDACTDISK


DRAGOBJ
!  #pts
     4
!           x            r          CDA
  0.40000E-01  0.60000E-01  0.40000E-01
  0.40000E-01  0.80000E-01  0.35000E-01
  0.40000E-01  0.10000      0.30000E-01
  0.40000E-01  0.12000      0.30000E-01
ENDDRAGOBJ

'''
    x_disc_stator         = dfdc_object.geometry.stator.percent_x_location  * length  
    header_2_text         = header_2_text_template.format(x_disc_stator ) 
    geometry.write(header_2_text)     
    return  

def make_hub_text(dfdc_object,geometry):  
    """This function writes the rotor using the template required for the DFDC executable to read 
    """ 
    duct_header = \
'''GEOM
FatDuct + CB test case
''' 
    geometry.write(duct_header) 
    if len(dfdc_object.geometry.hub_airfoil) > 0:
        if type(dfdc_object.geometry.hub_airfoil) == RCAIDE.Library.Components.Airfoils.NACA_4_Series_Airfoil:
            NACA_Code =  dfdc_object.geometry.hub_airfoil.NACA_4_Series_code
            airfoil_geometry_data = compute_naca_4series(NACA_Code)    
        else: 
            airfoil_filename      = dfdc_object.geometry.hub_airfoil.coordinate_file
            airfoil_geometry_data = import_airfoil_geometry(airfoil_filename)
    else:  
        airfoil_geometry_data = compute_naca_4series('0015') 
    dim = len(airfoil_geometry_data.x_upper_surface)
    x_coords =  airfoil_geometry_data.x_upper_surface[::-1]
    y_coords =  airfoil_geometry_data.y_upper_surface[::-1]
     
    for i in range(dim):
        x_coord  = x_coords[i]* (dfdc_object.geometry.hub_radius / max(y_coords))
        y_coord  = y_coords[i]* (dfdc_object.geometry.hub_radius / max(y_coords))
        case_text = '     ' + format(x_coord, '.6f')+ "    " + format(y_coord, '.6f') + "\n" 
        geometry.write(case_text)
              
    return


def make_separator_text(dfdc_object,geometry):  
    """This function writes the operating conditions using the template required for the DFDC executable to read 
    """      
    separator_text = \
'''  999.0 999.0
'''
    geometry.write(separator_text) 
    return


def make_duct_text(dfdc_object,geometry):  
    """This function writes the operating conditions using the template required for the DFDC executable to read
 
    """      
    if len(dfdc_object.geometry.duct_airfoil) > 0: 
        if type(dfdc_object.geometry.duct_airfoil) == RCAIDE.Library.Components.Airfoils.NACA_4_Series_Airfoil:
            NACA_Code =  dfdc_object.geometry.duct_airfoil.NACA_4_Series_code
            airfoil_geometry_data = compute_naca_4series(NACA_Code)    
        else: 
            airfoil_filename      = dfdc_object.geometry.duct_airfoil.coordinate_file
            airfoil_geometry_data = import_airfoil_geometry(airfoil_filename)
    else:  
        airfoil_geometry_data= compute_naca_4series('2208')    
    dim = len(airfoil_geometry_data.x_coordinates)

    x_coords  = airfoil_geometry_data.x_coordinates*dfdc_object.geometry.length 
    y_coords  = airfoil_geometry_data.y_coordinates*dfdc_object.geometry.length     
    offset    = abs(min(y_coords)) + dfdc_object.geometry.tip_radius
    for i in range(dim):  
        case_text = '     ' + format(x_coords[i], '.6f')+ "    " + format(y_coords[i] + offset, '.6f') + "\n" 
        geometry.write(case_text) 
    end_text = \
'''ENDGEOM
'''               
    geometry.write(end_text) 
    return  
 



# RCAIDE/Library/Methods/Weights/Correlation_Buildups/FLOPS/compute_operating_items_weight.py
# 
# 
# Created:  Sep 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE 
from RCAIDE.Framework.Core    import Units ,  Data

# python imports 
import  numpy as  np
 
# ----------------------------------------------------------------------------------------------------------------------
#  Operating Items Weight 
# ----------------------------------------------------------------------------------------------------------------------
def compute_operating_items_weight(vehicle):
    """ Calculate the weight of operating items, including:
        - crew
        - baggage
        - unusable fuel
        - engine oil
        - passenger service
        - ammunition and non-fixed weapons
        - cargo containers

        Assumptions:

        Source:
            http://aerodesign.stanford.edu/aircraftdesign/AircraftDesign.html

        Inputs:
            vehicle - data dictionary with vehicle properties                   [dimensionless]
                -.passengers: number of passengers
                -.systems.accessories: type of aircraft (short-range, commuter
                                                        medium-range, long-range,
                                                        sst, cargo)

        Outputs:
            output - data dictionary with weights                               [kilograms]
                    - output.oper_items: unusable fuel, engine oil, passenger service weight and cargo containers
                    - output.flight_crew: flight crew weight
                    - output.flight_attendants: flight attendants weight
                    - output.total: total operating items weight

        Properties Used:
            N/A
    """
    num_seats   = vehicle.passengers
    ac_type     = vehicle.systems.accessories
    if ac_type   == "short-range":  # short-range domestic, austere accomodations
        operitems_wt = 17.0 * num_seats * Units.lb
    elif ac_type == "medium-range":  # medium-range domestic
        operitems_wt = 28.0 * num_seats * Units.lb
    elif ac_type == "long-range":  # long-range overwater
        operitems_wt = 28.0 * num_seats * Units.lb
    elif ac_type == "business":  # business jet
        operitems_wt = 28.0 * num_seats * Units.lb
    elif ac_type == "cargo":  # all cargo
        operitems_wt = 56.0 * Units.lb
    elif ac_type == "commuter":  # commuter
        operitems_wt = 17.0 * num_seats * Units.lb
    elif ac_type == "sst":  # sst
        operitems_wt = 40.0 * num_seats * Units.lb
    else:
        operitems_wt = 28.0 * num_seats * Units.lb

    if vehicle.passengers >= 150:
        flight_crew = 3  # FLOPS: NFLCR
    else:
        flight_crew = 2

    if vehicle.passengers < 51:
        flight_attendants = 1  # FLOPS: NSTU
    else:
        flight_attendants = 1 + np.floor(vehicle.passengers / 40.)

    W_flight_attendants = flight_attendants * (170 + 40)  # FLOPS: WSTUAB
    W_flight_crew = flight_crew * (190 + 50)  # FLOPS: WFLCRB

    output                           = Data()
    output.misc = operitems_wt
    output.flight_crew               = W_flight_crew * Units.lbs
    output.flight_attendants         = W_flight_attendants * Units.lbs
    output.total                     = output.misc + output.flight_crew + \
                                       output.flight_attendants
    return output



# RCAIDE/Library/Plots/Energy/plot_battery_cell_conditions.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  

from RCAIDE.Framework.Core import Units
from RCAIDE.Library.Plots.Common import set_axes, plot_style
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import numpy as np 

# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ----------------------------------------------------------------------------------------------------------------------   
def plot_battery_cell_conditions(results,
                                  save_figure = False,
                                  show_legend = True,
                                  save_filename = "Battery_Cell_Conditions_",
                                  file_type = ".png",
                                  width = 11, height = 7):
    """Plots the cell-level conditions of the battery throughout flight.

    Assumptions:
    None

    Source:
    None

    Inputs:
    results.segments.conditions.
        freestream.altitude
        weights.total_mass
        weights.vehicle_mass_rate
        frames.body.thrust_force_vector

    Outputs: 
    Plots

    Properties Used:
    N/A	
    """ 
    
    # get plotting style 
    ps      = plot_style()  

    parameters = {'axes.labelsize': ps.axis_font_size,
                  'xtick.labelsize': ps.axis_font_size,
                  'ytick.labelsize': ps.axis_font_size,
                  'axes.titlesize': ps.title_font_size}
    plt.rcParams.update(parameters)
     

    fig = plt.figure(save_filename)
    fig.set_size_inches(width,height)
    # get line colors for plots 
    line_colors   = cm.inferno(np.linspace(0,0.9,len(results.segments)))      
    axis_1 = plt.subplot(3,2,1)
    axis_2 = plt.subplot(3,2,2) 
    axis_3 = plt.subplot(3,2,3) 
    axis_4 = plt.subplot(3,2,4)
    axis_5 = plt.subplot(3,2,5) 
    axis_6 = plt.subplot(3,2,6) 
     
    for network in results.segments[0].analyses.energy.vehicle.networks: 
        busses  = network.busses 
        for bus_i, bus in enumerate(busses):
            for b_i, battery in enumerate(bus.battery_modules):
                if b_i == 0 or bus.identical_battery_modules == False:
                    for i in range(len(results.segments)):  
                        time    = results.segments[i].conditions.frames.inertial.time[:,0] / Units.min    
                        battery_conditions  = results.segments[i].conditions.energy[bus.tag].battery_modules[battery.tag]    
                        cell_power          = battery_conditions.cell.power[:,0]
                        cell_energy         = battery_conditions.cell.energy[:,0]
                        cell_volts          = battery_conditions.cell.voltage_under_load[:,0]
                        cell_volts_oc       = battery_conditions.cell.voltage_open_circuit[:,0]
                        cell_current        = battery_conditions.cell.current[:,0]
                        cell_SOC            = battery_conditions.cell.state_of_charge[:,0]   
                        cell_temperature    = battery_conditions.cell.temperature[:,0]  
                
                        segment_tag  = results.segments[i].tag
                        segment_name = segment_tag.replace('_', ' ') 
    
                        if b_i == 0 and i ==0:                    
                            axis_1.plot(time, cell_SOC, color = line_colors[i], marker = ps.markers[bus_i], linewidth = ps.line_width, label = battery.tag)
                        else:
                            axis_1.plot(time, cell_SOC, color = line_colors[i], marker = ps.markers[bus_i], linewidth = ps.line_width)
                        axis_1.set_ylabel(r'SOC')
                        axis_1.set_ylim([0,1.1])
                        set_axes(axis_1)     
                         
                        axis_2.plot(time, cell_energy/Units.Wh, color = line_colors[i], marker = ps.markers[bus_i], linewidth = ps.line_width)
                        axis_2.set_ylabel(r'Energy (W-hr)')
                        set_axes(axis_2) 
                 
                        axis_3.plot(time, cell_current, color = line_colors[i], marker = ps.markers[bus_i], linewidth = ps.line_width)
                        axis_3.set_ylabel(r'Current (A)')
                        set_axes(axis_3)  
                 
                        axis_4.plot(time, cell_power, color = line_colors[i], marker = ps.markers[bus_i], linewidth = ps.line_width)
                        axis_4.set_ylabel(r'Power (W)')
                        set_axes(axis_4)     
                         
                        axis_5.plot(time, cell_volts, color = line_colors[i], marker = ps.markers[bus_i], linewidth = ps.line_width) 
                        axis_5.set_ylabel(r'Voltage (V)')
                        set_axes(axis_5) 
                 
                        axis_6.plot(time, cell_temperature, color = line_colors[i], marker = ps.markers[bus_i], linewidth = ps.line_width)
                        axis_6.set_ylabel(r'Temperature, $\degree$C')
                        set_axes(axis_6)  
                  
    if show_legend:      
        leg =  fig.legend(bbox_to_anchor=(0.5, 0.95), loc='upper center', ncol = 4)   
    
    # Adjusting the sub-plots for legend 
    fig.tight_layout()
    fig.subplots_adjust(top=0.8) 
    
    # set title of plot 
    title_text   = 'Battery Cell Conditions'       
    fig.suptitle(title_text) 
    
    if save_figure:
        plt.savefig(save_filename + battery.tag + file_type)    
    return fig  


# RCAIDE/Library/Plots/Energy/plot_fuel_consumption.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
from RCAIDE.Framework.Core import Units
from RCAIDE.Library.Plots.Common import set_axes, plot_style
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import numpy as np 

# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ----------------------------------------------------------------------------------------------------------------------   
def plot_fuel_consumption(results,
                    save_figure = False,
                    show_legend=True,
                    save_filename = "Aircraft_Fuel_Burnt",
                    file_type = ".png",
                    width = 11, height = 7): 

    """This plots aircraft fuel usage
    
    
    Assumptions:
    None

    Source: 
  
    Inputs:
    results.segments.condtions.
        frames.inertial.time
        weights.fuel_mass
        weights.additional_fuel_mass
        weights.total_mass
        
    Outputs:
    Plots
    Properties Used:
    N/A	"""

    # get plotting style 
    ps      = plot_style()  

    parameters = {'axes.labelsize': ps.axis_font_size,
                  'xtick.labelsize': ps.axis_font_size,
                  'ytick.labelsize': ps.axis_font_size,
                  'axes.titlesize': ps.title_font_size}
    plt.rcParams.update(parameters)
     
    # get line colors for plots 
    line_colors   = cm.inferno(np.linspace(0,0.9,len(results.segments)))       
  
    fig = plt.figure(save_filename)
    fig.set_size_inches(width,height)

    prev_seg_fuel       = 0
    prev_seg_extra_fuel = 0
    total_fuel          = 0

    axes = plt.subplot(1,1,1)
            
    for i in range(len(results.segments)):

        segment  = results.segments[i]
        time     = segment.conditions.frames.inertial.time[:,0] / Units.min  
        segment_tag  =  results.segments[i].tag
        segment_name = segment_tag.replace('_', ' ')        

        if "has_additional_fuel" in segment.conditions.weights and segment.conditions.weights.has_additional_fuel == True:


            fuel     = segment.conditions.weights.fuel_mass[:,0]
            alt_fuel = segment.conditions.weights.additional_fuel_mass[:,0]

            if i == 0:

                plot_fuel     = np.negative(fuel)
                plot_alt_fuel = np.negative(alt_fuel)

                axes.plot( time , plot_fuel , 'ro-', marker = ps.markers[0], linewidth = ps.line_width , label = 'fuel')
                axes.plot( time , plot_alt_fuel , 'bo-', marker = ps.markers[0], linewidth = ps.line_width, label = 'additional fuel' )
                axes.plot( time , np.add(plot_fuel, plot_alt_fuel), 'go-', marker = ps.markers[0], linewidth = ps.line_width, label = 'total fuel' )

                axes.legend(loc='center right')   

            else:
                prev_seg_fuel       += results.segments[i-1].conditions.weights.fuel_mass[-1]
                prev_seg_extra_fuel += results.segments[i-1].conditions.weights.additional_fuel_mass[-1]

                current_fuel         = np.add(fuel, prev_seg_fuel)
                current_alt_fuel     = np.add(alt_fuel, prev_seg_extra_fuel)

                axes.plot( time , np.negative(current_fuel)  , 'ro-' , marker = ps.markers[0], linewidth = ps.line_width)
                axes.plot( time , np.negative(current_alt_fuel ), 'bo-', marker = ps.markers[0], linewidth = ps.line_width)
                axes.plot( time , np.negative(current_fuel + current_alt_fuel), 'go-', marker = ps.markers[0], linewidth = ps.line_width)

        else:
            
            initial_weight  = results.segments[0].conditions.weights.total_mass[:,0][0] 
            fuel            = results.segments[i].conditions.weights.total_mass[:,0]
            time            = results.segments[i].conditions.frames.inertial.time[:,0] / Units.min 
            total_fuel      = np.negative(results.segments[i].conditions.weights.total_mass[:,0] - initial_weight )
            axes.plot( time, total_fuel, color = line_colors[i], marker = ps.markers[0], linewidth = ps.line_width, label = segment_name)

    axes.set_ylabel('Fuel (kg)')
    axes.set_xlabel('Time (min)')

    set_axes(axes)  

    if show_legend:    
        leg =  fig.legend(bbox_to_anchor=(0.5, 0.95), loc='upper center', ncol = 4) 
        leg.set_title('Flight Segment', prop={'size': ps.legend_font_size, 'weight': 'heavy'})    
    
    # Adjusting the sub-plots for legend 
    fig.tight_layout()
    fig.subplots_adjust(top=0.8)
    
    # set title of plot 
    title_text    = 'Aircraft Fuel Burnt'      
    fig.suptitle(title_text) 

    if save_figure:
        plt.savefig(save_filename + file_type)  
        
    return fig 


# RCAIDE/Methods/Energy/Distributors/Electrical_Bus.py
# 
# 
# Created: Sep 2024, S. Shekar

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
import numpy as np


def compute_bus_conditions(bus, state, t_idx, delta_t): 
    """Computes the conditions of the bus based on the response of the battery modules
    
    Args:
        bus: Electrical bus object
        state: Current system state
        t_idx: Time index
        delta_t: Time step
    """  
    bus_conditions = state.conditions.energy[bus.tag]

    if bus.battery_module_electric_configuration == 'Series':
        bm_conditions                               = [bus_conditions.battery_modules[bm.tag] for bm in bus.battery_modules]
        bus_conditions.voltage_open_circuit[t_idx]  = sum(bm.voltage_open_circuit[t_idx] for bm in bm_conditions)
        bus_conditions.voltage_under_load[t_idx]    = sum(bm.voltage_under_load[t_idx] for bm in bm_conditions)
        bus_conditions.heat_energy_generated[t_idx] = sum(bm.heat_energy_generated[t_idx] for bm in bm_conditions)
        bus_conditions.efficiency[t_idx]            = (bus_conditions.power_draw[t_idx] + bus_conditions.heat_energy_generated[t_idx])/bus_conditions.power_draw[t_idx]
        if t_idx != state.numerics.number_of_control_points-1:  
            bm_conditions                              = [bus_conditions.battery_modules[bm.tag] for bm in bus.battery_modules]
            bus_conditions.temperature[t_idx+1]        = sum(bm.temperature[t_idx+1] for bm in bm_conditions)/ bus.number_of_battery_modules
            bus_conditions.energy[t_idx+1]             = sum(bm.energy[t_idx+1] for bm in bm_conditions)
            bus_conditions.state_of_charge[t_idx+1]    = bm_conditions[-1].state_of_charge[t_idx+1]

    elif bus.battery_module_electric_configuration == 'Parallel':
        bm_conditions                               = [bus_conditions.battery_modules[bm.tag] for bm in bus.battery_modules]
        bus_conditions.heat_energy_generated[t_idx] = sum(bm.heat_energy_generated[t_idx] for bm in bm_conditions)
        bus_conditions.voltage_open_circuit[t_idx]  = bm_conditions[-1].voltage_open_circuit[t_idx]
        bus_conditions.voltage_under_load[t_idx]    = bm_conditions[-1].voltage_under_load[t_idx]             
        bus_conditions.efficiency[t_idx]            = (bus_conditions.power_draw[t_idx] +  bus_conditions.heat_energy_generated[t_idx])/bus_conditions.power_draw[t_idx]
        if t_idx != state.numerics.number_of_control_points-1:  
            bus_conditions.heat_energy_generated[t_idx] = sum(bm.heat_energy_generated[t_idx] for bm in bm_conditions)
            bus_conditions.temperature[t_idx+1]         = sum(bm.temperature[t_idx+1] for bm in bm_conditions)/bus.number_of_battery_modules
            bus_conditions.energy[t_idx+1]              = sum(bm.energy[t_idx+1] for bm in bm_conditions)
            bus_conditions.state_of_charge[t_idx+1]     = bm_conditions[-1].cell.state_of_charge[t_idx+1]
    
    if t_idx != state.numerics.number_of_control_points-1:  
        # Handle fully charged state
        if state.conditions.energy.recharging and np.float16(bus_conditions.state_of_charge[t_idx+1]) == 1:
            bus_conditions.charging_current[t_idx+1] = 0
            bus_conditions.power_draw[t_idx+1]       = 0
            bus_conditions.current_draw[t_idx+1]     = 0
    return


# RCAIDE/Library/Methods/Aerodynamics/Common/Drag/supersonic_wave_drag_volume_raymer.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
import RCAIDE
from RCAIDE.Framework.Core                    import    Units 
from RCAIDE.Library.Components.Wings          import Main_Wing
from RCAIDE.Library.Methods.Utilities         import Cubic_Spline_Blender   
from RCAIDE.Library.Methods.Geometry.Planform.convert_sweep import convert_sweep 

# package imports
import numpy as np

# ---------------------------------------------------------------------------------------------------------------------- 
#  Supersonic Wave Drag Due to Volume - Raymer Method 
# ---------------------------------------------------------------------------------------------------------------------- 
def supersonic_wave_drag_volume_raymer(vehicle,mach,scaling_factor):
    """Computes the volume drag

    Assumptions:
    Basic fit

    Source:
    D. Raymer, Aircraft Design: A Conceptual Approach, Fifth Ed. pg. 448-449

    Args:
    vehicle.
      wings.main_wing.sweeps.leading_edge [rad]
      total_length                        [m]
      maximum_cross_sectional_area        [m^2]
      reference_area                      [m^2]
      
    Returns:
    vehicle_wave_drag                     [Unitless] 
    """
    L =  0
    for wing in vehicle.wings:
        if isinstance(wing,Main_Wing):
            main_wing = wing
            L  =  wing.chords.root
    
    main_wing = vehicle.wings.main_wing
    # estimation of leading edge sweep if not defined 
    if main_wing.sweeps.leading_edge == None:                           
        main_wing.sweeps.leading_edge  = convert_sweep(main_wing,old_ref_chord_fraction = 0.25 ,new_ref_chord_fraction = 0.0) 
     
    for fuselage in vehicle.fuselages:
        if type(fuselage) == RCAIDE.Library.Components.Fuselages.Blended_Wing_Body_Fuselage:
            pass
        else:
            L =  np.maximum(L, fuselage.lengths.total)
        
    LE_sweep = main_wing.sweeps.leading_edge / Units.deg
    Ae       = vehicle.maximum_cross_sectional_area
    S        = vehicle.reference_area
    
    # Compute sears-hack D/q
    Dq_SH = 9*np.pi/2*(Ae/L)*(Ae/L)
    
    spline = Cubic_Spline_Blender(1.2,1.3)
    h00    = lambda M:spline.compute(M)    
    
    # Compute full vehicle D/q
    Dq_vehicle           = np.zeros_like(mach)
    Dq_vehicle_simpified = np.zeros_like(mach)
    
    Dq_vehicle[mach>=1.2] = scaling_factor*(1-0.2*(mach[mach>=1.2]-1.2)**0.57*(1-np.pi*LE_sweep**.77/100))*Dq_SH
    Dq_vehicle_simpified  = scaling_factor*Dq_SH
    
    Dq_vehicle = Dq_vehicle_simpified*h00(mach) + Dq_vehicle*(1-h00(mach))
    
    CD_c_vehicle = Dq_vehicle/S
    
    return CD_c_vehicle


# RCAIDE/Library/Components/Thermal_Management/Batteries/Air_Cooled.py
# 
# Created:  Jul 2023, M. Clarke
# Modified: Aug 2024, S. Shekar 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
from RCAIDE.Library.Components import Component  
from RCAIDE.Library.Methods.Thermal_Management.Batteries.Air_Cooled import append_air_cooled_conditions, air_cooled_performance, append_air_cooled_segment_conditions
from RCAIDE.Library.Attributes.Gases import Air
from RCAIDE.Library.Plots.Thermal_Management.plot_air_cooled_conditions import plot_air_cooled_conditions

# ----------------------------------------------------------------------------------------------------------------------
#  Air_Cooled
# ----------------------------------------------------------------------------------------------------------------------
class Air_Cooled(Component):
    """
    A class representing an air-cooled thermal management system for battery packs using 
    direct convection cooling.

    Attributes
    ----------
    tag : str
        Unique identifier for the cooling system, defaults to 'air_cooled_heat_acquisition'
        
    cooling_fluid : Air
        Air properties for cooling calculations, defaults to standard Air()
        
    cooling_fluid.flowspeed : float
        Air flow velocity through the cooling channels, defaults to 0.01
        
    convective_heat_transfer_coefficient : float
        Heat transfer coefficient between air and battery surface, defaults to 35.0
        
    heat_transfer_efficiency : float
        Overall efficiency of the heat transfer process, defaults to 1.0

    Notes
    -----
    The air-cooled system provides direct convective cooling for battery thermal 
    management. It includes functionality for:
    
    * Operating condition analysis
    * Thermal performance calculation
    * Condition plotting and visualization

    **Definitions**

    'Convective Heat Transfer Coefficient'
        Rate of heat transfer between the battery surface and cooling air
        
    'Heat Transfer Efficiency'
        Ratio of actual to theoretical maximum heat transfer

    See Also
    --------
    RCAIDE.Library.Components.Thermal_Management.Batteries.Liquid_Cooled_Wavy_Channel
        Alternative cooling approach using liquid coolant
    RCAIDE.Library.Components.Thermal_Management.Batteries.Cryocooler
        Alternative cooling approach for extreme temperatures
    """
    
    def __defaults__(self):
        """
        Sets default values for the air cooling system attributes.
        """                 
        self.tag                                  = 'air_cooled_heat_acquisition'
        self.cooling_fluid                        = Air()    
        self.cooling_fluid.flowspeed              = 0.01                                          
        self.convective_heat_transfer_coefficient = 35.     
        self.heat_transfer_efficiency             = 1.0      
   
    def append_operating_conditions(self, segment, coolant_line):
        """
        Adds operating conditions for the cooling system to a mission segment.

        Parameters
        ----------
        segment : Data
            Mission segment to which conditions are being added
        coolant_line : Data
            Cooling system flow path information
        """
        append_air_cooled_conditions(self, segment, coolant_line)
        return

    def append_segment_conditions(self, segment, coolant_line, conditions):
        """
        Adds specific segment conditions to the cooling system analysis.

        Parameters
        ----------
        segment : Data
            Mission segment being analyzed
        coolant_line : Data
            Cooling system flow path information
        conditions : Data
            Operating conditions for the segment
        """
        append_air_cooled_segment_conditions(self, segment, coolant_line, conditions)
        return
    
    def compute_thermal_performance(self, battery, bus, coolant_line, Q_heat_gen, T_cell, state, delta_t, t_idx): 
        """
        Calculates thermal performance of the air cooling system.

        Parameters
        ----------
        battery : Data
            Battery pack information
        bus : Data
            Electrical bus data
        coolant_line : Data
            Cooling system flow path information
        Q_heat_gen : float
            Heat generation rate
        T_cell : float
            Current cell temperature
        state : Data
            Current system state
        delta_t : float
            Time step size
        t_idx : int
            Time index in the simulation

        Returns
        -------
        float
            Updated battery temperature
        """
        T_battery_current = air_cooled_performance(self, battery, bus, coolant_line, 
                                                 Q_heat_gen, T_cell, state, delta_t, t_idx)
        return T_battery_current

    def plot_operating_conditions(self, results, coolant_line, save_filename, save_figure, 
                                show_legend, file_type, width, height):
        """
        Creates visualization plots of the cooling system operating conditions.

        Parameters
        ----------
        results : Data
            Simulation results data
        coolant_line : Data
            Cooling system flow path information
        save_filename : str
            Path for saving the plot
        save_figure : bool
            Flag to save the figure
        show_legend : bool
            Flag to display plot legend
        file_type : str
            Output file format
        width : float
            Plot width
        height : float
            Plot height
        """
        plot_air_cooled_conditions(self, results, coolant_line, save_filename, 
                                 save_figure, show_legend, file_type, width, height)
        return
        



# RCAIDE/Library/Compoments/Energy/Sources/Batteries/Lithium_Ion_LiNiMnCoO2_18650.py
# 
# 
# Created:  Mar 2024, M. Clarke
 
# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
# RCAIDE imports
import RCAIDE
from RCAIDE.Framework.Core                                            import Units , Data
from .Generic_Battery_Module                                          import Generic_Battery_Module   
from RCAIDE.Library.Methods.Energy.Sources.Batteries.Lithium_Ion_NMC  import *
# package imports 
import numpy as np
import os 
from scipy.interpolate  import RegularGridInterpolator 

# ----------------------------------------------------------------------------------------------------------------------
#  Lithium_Ion_NMC
# ----------------------------------------------------------------------------------------------------------------------  
class Lithium_Ion_NMC(Generic_Battery_Module):
    """
    Class for modeling 18650-format lithium nickel manganese cobalt oxide batteries
    
    Attributes
    ----------
    tag : str
        Identifier for the battery module (default: 'lithium_ion_nmc')
        
    maximum_energy : float
        Maximum energy storage capacity [J] (default: 0.0)
        
    maximum_power : float
        Maximum power output [W] (default: 0.0)
        
    maximum_voltage : float
        Maximum voltage output [V] (default: 0.0)
        
    cell : Data
        Cell-specific properties
        
        - chemistry : str
            Battery chemistry type (default: 'LiNiMnCoO2')
        - diameter : float
            Cell diameter [m] (default: 0.0185)
        - height : float
            Cell height [m] (default: 0.0653)
        - mass : float
            Cell mass [kg] (default: 0.048)
        - surface_area : float
            Total cell surface area [m^2]
        - volume : float
            Cell volume [m^3]
        - density : float
            Cell density [kg/m^3]
        - electrode_area : float
            Active electrode area [m^2] (default: 0.0342)
        - maximum_voltage : float
            Maximum cell voltage [V] (default: 4.2)
        - nominal_capacity : float
            Rated capacity [Ah] (default: 3.8)
        - nominal_voltage : float
            Nominal operating voltage [V] (default: 3.6)
        - charging_voltage : float
            Charging voltage [V] (default: nominal_voltage)
        - resistance : float
            Internal resistance [Ohms] (default: 0.025)
        - specific_heat_capacity : float
            Cell specific heat [J/kgK] (default: 1108)
        - radial_thermal_conductivity : float
            Radial thermal conductivity [W/mK] (default: 0.4)
        - axial_thermal_conductivity : float
            Axial thermal conductivity [W/mK] (default: 32.2)
        - discharge_performance_map : RegularGridInterpolator
            Interpolator for voltage vs discharge characteristics

    Notes
    -----
    The NMC cell model includes detailed thermal and electrical characteristics
    based on 18650-format cells. The model includes forced air cooling assumptions
    with a convective heat transfer coefficient for 35 m/s airflow.

    References
    ----------
    [1] Jeon, D. H., & Baek, S. M. (2011). Thermal modeling of cylindrical 
        lithium ion battery during discharge cycle. Energy Conversion and Management,
        52(8-9), 2973-2981.
        
    [2] Yang, S., et al. (2019). A Review of Lithium-Ion Battery Thermal Management 
        System Strategies and the Evaluate Criteria. Int. J. Electrochem. Sci, 14,
        6077-6107.
        
    [3] Muenzel, V., et al. (2015). A comparative testing study of commercial
        18650-format lithium-ion battery cells. Journal of The Electrochemical
        Society, 162(8), A1592.

    See Also
    --------
    RCAIDE.Library.Components.Energy.Sources.Battery_Modules.Generic_Battery_Module
        Base battery module class
    """       
    
    def __defaults__(self):   
        """This sets the default values.
    
        Assumptions:
            Convective Thermal Conductivity Coefficient corresponds to forced
            air cooling in 35 m/s air 
        
        Source: 
            convective  heat transfer coefficient, h 
            Jeon, Dong Hyup, and Seung Man Baek. "Thermal modeling of cylindrical 
            lithium ion battery during discharge cycle." Energy Conversion and Management
            52.8-9 (2011): 2973-2981.
            
            thermal conductivity, k 
            Yang, Shuting, et al. "A Review of Lithium-Ion Battery Thermal Management 
            System Strategies and the Evaluate Criteria." Int. J. Electrochem. Sci 14
            (2019): 6077-6107.
            
            specific heat capacity, Cp
            (axial and radial)
            Yang, Shuting, et al. "A Review of Lithium-Ion Battery Thermal Management 
            System Strategies and the Evaluate Criteria." Int. J. Electrochem. Sci 14
            (2019): 6077-6107.
            
            # Electrode Area
            Muenzel, Valentin, et al. "A comparative testing study of commercial
            18650-format lithium-ion battery cells." Journal of The Electrochemical
            Society 162.8 (2015): A1592.
        
        """
        # ----------------------------------------------------------------------------------------------------------------------
        #  Module Level Properties
        # ----------------------------------------------------------------------------------------------------------------------
        
        self.tag                                         = 'lithium_ion_nmc'
        self.maximum_energy                              = 0.0
        self.maximum_power                               = 0.0
        self.maximum_voltage                             = 0.0  
         
        # ----------------------------------------------------------------------------------------------------------------------
        #  Cell Level Properties
        # ----------------------------------------------------------------------------------------------------------------------        
        self.cell.chemistry                   = 'LiNiMnCoO2'
        self.cell.diameter                    = 0.0185                                                                            # [m]
        self.cell.height                      = 0.0653                                                                            # [m]
        self.cell.mass                        = 0.048 * Units.kg                                                                  # [kg]
        self.cell.surface_area                = (np.pi*self.cell.height*self.cell.diameter) + (0.5*np.pi*self.cell.diameter**2)  # [m^2]
        self.cell.volume                      = np.pi*(0.5*self.cell.diameter)**2*self.cell.height 
        self.cell.density                     = self.cell.mass/self.cell.volume                                                  # [kg/m^3]  
        self.cell.electrode_area              = 0.0342                                                                           # [m^2] 
                                                                                                                           
        self.cell.maximum_voltage             = 4.2                                                                              # [V]
        self.cell.nominal_capacity            = 3.8                                                                             # [Amp-Hrs]
        self.cell.nominal_voltage             = 3.6                                                                              # [V] 
        self.cell.charging_voltage            = self.cell.nominal_voltage                                                        # [V] 
        
        self.cell.watt_hour_rating            = self.cell.nominal_capacity  * self.cell.nominal_voltage                          # [Watt-hours]      
        self.cell.specific_energy             = self.cell.watt_hour_rating*Units.Wh/self.cell.mass                               # [J/kg]
        self.cell.specific_power              = self.cell.specific_energy/self.cell.nominal_capacity                             # [W/kg]   
        self.cell.resistance                  = 0.025                                                                            # [Ohms] 
                                                            
        self.cell.specific_heat_capacity      = 1108                                                                             # [J/kgK]    
        self.cell.radial_thermal_conductivity = 0.4                                                                              # [J/kgK]  
        self.cell.axial_thermal_conductivity  = 32.2                                                                             # [J/kgK] # estimated
    
                                              
        battery_raw_data                      = load_battery_results()                                                   
        self.cell.discharge_performance_map   = create_discharge_performance_map(battery_raw_data)  

        return  
    
    def energy_calc(self,state,bus,coolant_lines, t_idx, delta_t): 
        """
        Computes the state of the NMC battery cell
        
        This method calculates the battery's electrical performance and thermal
        behavior during operation, including voltage, current, power, and 
        temperature distributions.

        Parameters
        ----------
        state : Data
            Current system state containing:
            - Temperature distributions
            - Power demands
            - Operating conditions
            
        bus : Component
            Connected electrical bus containing:
            - Voltage requirements
            - Power requirements
            - Load characteristics
            
        coolant_lines : Component
            Thermal management system containing:
            - Coolant properties
            - Flow conditions
            - Heat exchanger parameters
            
        t_idx : int
            Current time index in the simulation
            
        delta_t : float
            Time step size [s]

        Returns
        -------
        stored_results_flag : bool
            Flag indicating if results were stored for future reuse
            
        stored_battery_tag : str
            Identifier for stored battery state data

        Notes
        -----
        The calculation includes:
        - Voltage and current based on load demand
        - Heat generation from internal resistance
        - Thermal distribution with cooling effects
        - State of charge tracking
        """        
        stored_results_flag, stored_battery_tag =  compute_nmc_cell_performance(self,state,bus,coolant_lines, t_idx,delta_t) 
        
        return stored_results_flag, stored_battery_tag
    
    def reuse_stored_data(self,state,bus,coolant_lines, t_idx, delta_t,stored_results_flag, stored_battery_tag):
        reuse_stored_nmc_cell_data(self,state,bus,coolant_lines, t_idx, delta_t,stored_results_flag, stored_battery_tag)
        return 
    
    def update_battery_age(self,segment,battery_conditions,increment_battery_age_by_one_day = False):  
        """
        Updates battery aging parameters based on usage and environmental conditions
        
        This method tracks battery degradation by considering factors such as:
        cycle count, depth of discharge, temperature exposure, and calendar aging.

        Parameters
        ----------
        segment : Segment
            Flight segment containing:
            - Duration
            - Operating conditions
            - Power profile
            
        battery_conditions : Data
            Battery state data including:
            - Temperature history
            - Current rates
            - State of charge history
            
        increment_battery_age_by_one_day : bool, optional
            Flag to increment calendar age (default: False)

        Notes
        -----
        The aging model accounts for:
        - Capacity fade from cycling
        - Calendar aging effects
        - Temperature-dependent degradation
        - Current rate impacts
        """        
        update_nmc_cell_age(self,segment,battery_conditions,increment_battery_age_by_one_day) 
        
        return  

def create_discharge_performance_map(raw_data):
    """
    Creates discharge and charge response surface for a LiNiMnCoO2 battery cell   

    Parameters
    ----------
    raw_data : Data
        Container with experimental battery data including:
        - Voltage : array
            Discharge voltage curves at different currents and temperatures
        - Temperature : array
            Cell temperature profiles during discharge

    Returns
    -------
    battery_data : Data
        Container with interpolation functions:
        - Voltage : RegularGridInterpolator
            Predicts voltage based on [current, temperature, SOC]
        - Temperature : RegularGridInterpolator
            Predicts cell temperature based on [current, temperature, SOC]

    Notes
    -----
    The function creates 3D interpolations for:
    - Voltage as function of current (0-8A), temperature (0-50°C), and SOC (0-1)
    - Temperature rise as function of same parameters
    
    Uses regular grid interpolation for smooth predictions across the operating space.
    """   
    # Process raw data   
    processed_data = Data() 
    processed_data.Voltage        = np.zeros((5,6,15,2)) # current , operating temperature , state_of_charge vs voltage      
    processed_data.Temperature    = np.zeros((5,6,15,2)) # current , operating temperature , state_of_charge vs temperature 

    # Reshape  Data          
    raw_data.Voltage 
    for i, Amps in enumerate(raw_data.Voltage):
        for j , Deg in enumerate(Amps):
            min_x    = 0 
            max_x    = max(Deg[:,0])
            x        = np.linspace(min_x,max_x,15)
            y        = np.interp(x,Deg[:,0],Deg[:,1])
            vec      = np.zeros((15,2))
            vec[:,0] = x/max_x
            vec[:,1] = y
            processed_data.Voltage[i,j,:,:]= vec   

    for i, Amps in enumerate(raw_data.Temperature):
        for j , Deg in enumerate(Amps):
            min_x    = 0   
            max_x    = max(Deg[:,0])
            x        = np.linspace(min_x,max_x,15)
            y        = np.interp(x,Deg[:,0],Deg[:,1])
            vec      = np.zeros((15,2))
            vec[:,0] = x/max_x
            vec[:,1] = y
            processed_data.Temperature[i,j,:,:]= vec  
    
    # Create performance maps  
    battery_data             = Data() 
    amps                    = np.linspace(0, 8, 5)
    temp                    = np.linspace(0, 50, 6) +  272.65  # Convert to Kelvin
    SOC                     = np.linspace(0, 1, 15)
    battery_data.Voltage     = RegularGridInterpolator((amps, temp, SOC), processed_data.Voltage,bounds_error=False,fill_value=None)
    battery_data.Temperature = RegularGridInterpolator((amps, temp, SOC), processed_data.Temperature,bounds_error=False,fill_value=None) 
     
    return battery_data  

def load_battery_results(): 
    '''Load experimental raw data of NMC cells 
        
       Assumptions:
           Ideal gas
           
       Source:
           Automotive Industrial Systems Company of Panasonic Group, Technical Information of 
           NCR18650G, URL https://www.imrbatteries.com/content/panasonic_ncr18650g.pdf
    
       Args: 
           None
           
       Returns:
           battery_data: raw data from battery   [unitless]
    '''    
    ospath    = os.path.abspath(__file__)
    separator = os.path.sep
    rel_path  = os.path.dirname(ospath) + separator     
    return RCAIDE.load(rel_path+ 'NMC_Raw_Data.res')


# RCAIDE/Library/Methods/Thermal_Management/Heat_Exchangers/Cross_Flow_Heat_Exchanger/cross_flow_heat_exchanger_geometry_setup.py
#
# Created: Jun 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE imports 
import RCAIDE
from copy import  deepcopy

# ----------------------------------------------------------------------------------------------------------------------  
#  Cross Flow Heat Exchanger Geometry Setup 
# ----------------------------------------------------------------------------------------------------------------------   
def cross_flow_heat_exchanger_geometry_setup(HEX,coolant_line_base): 
    """ Modifies geometry of Cross Flow Heat Exchanger  
          
          Inputs:  
             nexus     - RCAIDE optmization framework with Wavy Channel geometry data structure [None]
              
          Outputs:   
             procedure - optimization methodology                                       
              
          Assumptions: 
             N/A 
        
          Source:
             None
    """            
    vehicle                = RCAIDE.Vehicle()  
    net                    = RCAIDE.Framework.Networks.Electric()             
    coolant_line           = deepcopy(coolant_line_base) 

    # To create an empty container for the Heat Exchanger
    RCAIDE.Library.Components.Thermal_Management.Batteries.Liquid_Cooled_Wavy_Channel(coolant_line)
    
    HEX.coolant_temperature_of_hot_fluid                  = 323
    HEX.design_heat_removed                               = 100000 
    coolant_line.heat_exchangers.append(HEX)     

    net.coolant_lines.append(coolant_line) 
    vehicle.append_energy_network(net) 
    

    
    configs                             = RCAIDE.Library.Components.Configs.Config.Container()
    base_config                         = RCAIDE.Library.Components.Configs.Config(vehicle)  
    config                              = RCAIDE.Library.Components.Configs.Config(base_config)
    config.tag                          = 'optimized'  
    configs.append(config)         
    return configs 


# RCAIDE/Library/Methods/Propulsors/Converters/fan/append_fan_conditions.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke  

from RCAIDE.Framework.Mission.Common     import   Conditions

# ---------------------------------------------------------------------------------------------------------------------- 
#  append_fan_conditions
# ----------------------------------------------------------------------------------------------------------------------    
def append_fan_conditions(fan,segment,propulsor_conditions): 
    propulsor_conditions[fan.tag]                              = Conditions() 
    propulsor_conditions[fan.tag].inputs                       = Conditions() 
    propulsor_conditions[fan.tag].outputs                      = Conditions() 
    return 